(* Title:  BaseDefinitions/celem-7.sml
   Author: Walther Neuper
   (c) due to copyright terms

Here is a minimum of code required for Know_Store.thy.
For main code see structure CAS_Cmd.
*)
signature COMPUTER_ALGEBRA_SYSTEM__COMMAND_DEF =
sig
  type input_pos
  val parse: Proof.context -> input_pos -> term option

  type generate_fn
  type T
(*type cas_elem*)
  val equal: T * T -> bool
(*val cas_eq: T * T -> bool*)
  val set_data: generate_fn -> theory -> theory
  val the_data: theory -> generate_fn
end

(**)
structure CAS_Def(**): COMPUTER_ALGEBRA_SYSTEM__COMMAND_DEF(**) =
struct
(**)

type input_pos = ((*TermC.as_*)string * Position.T) option;

fun parse ctxt cas =
  (case cas of
    SOME (s, pos) =>
      SOME (Syntax.read_term ctxt s
              handle ERROR msg => error (msg ^ Position.here pos))
  | NONE => NONE)


(* association list with cas-commands, for generating a complete calc-head *)
type generate_fn = 
  (term list ->         (* the arguments of the cas-command, eg. (x+1=2, x) *)
    (term *             (* description of an element                        *)
      term list)        (* value of the element (always put into a list)    *)
  list)                 (* of elements in the formalization                 *)
type T = 
  (term *               (* cas-command, eg. 'solve'                         *)
    (References_Def.T * (* theory, problem, method                          *)
    generate_fn))
fun equal ((t1, (_, _)) : T, (t2, (_, _)) : T) = t1 = t2


(* theory data for Isar command 'cas' *)

structure Data = Theory_Data
(
  type T = generate_fn option;
  val empty = NONE;
  fun merge _ = NONE;
);

val set_data = Data.put o SOME;
val the_data = the o Data.get;

(**)end(*struct*)
