(*  Title:      src/Tools/isac/BridgeJEdit/Calculation.thy
    Author:     Walther Neuper, JKU Linz

Outer syntax for Isabelle/Isac's Calculation interactive via Isabelle/PIDE.
Second trial following Makarius' definition of "problem" already existing in this repository.
*)

theory Calculation
  imports
  (**)"$ISABELLE_ISAC/Knowledge/Build_Thydata"   (*remove after devel.of BridgeJEdit*)
  (**)"$ISABELLE_ISAC/MathEngine/MathEngine"
  keywords "Example" :: thy_decl
    and "Specification" "Model" "References" "Solution" 
begin

ML_file "parseC.sml"
ML_file "preliminary.sml"

section \<open>Outer syntax command \<^theory_text>\<open>Example\<close>\<close>

text \<open>
  Stepwise development starts at changeset
  \<^url>\<open>https://hg.risc.uni-linz.ac.at/wneuper/isa/rev/9d98791b4080\<close>
  and can be inspected following the subsequent changesets..
\<close>

ML \<open>
\<close> ML \<open>
\<close> ML \<open>
structure Data = Theory_Data
(
  type T = Ctree.ctree option;
  val empty = NONE;
  fun merge _ = NONE;
);

val set_data = Data.put o SOME;
fun the_data thy = 
  case Data.get thy of NONE => Ctree.EmptyPtree
  | SOME ctree => ctree;
\<close> ML \<open>
\<close> text \<open>
  A provisional Example collection.
  
  For notes from ISAC's discontinued Java front-end see
  \<^url>\<open>https://isac.miraheze.org/wiki/Extend_ISAC_Knowledge#Add_an_example_in_XML_format\<close>
\<close> ML \<open>
\<close> ML \<open>
val demo_example = ("The efficiency of an electrical coil depends on the mass of the kernel." ^
  "Given is a kernel with cross-sectional area determined by two rectangles" ^
  "of same shape as shown in the figure." ^
  "Given a radius r = 7, determine the length u and width v of the rectangles" ^
  "such that the cross-sectional area A (and thus the mass of the kernel) is a" ^
  "maximum. + Figure", 
   ([
   (*Problem model:*)
     "Constants [r = (7::real)]", "Maximum A", "AdditionalValues [u, v]",
     "Extremum (A = 2 * u * v - u \<up> 2)",
     "SideConditions [((u::real) / 2) \<up> 2 + (2 / v) \<up> 2 = r \<up> 2]",
     "SideConditions [((u::real) / 2) \<up> 2 + (2 / v) \<up> 2 = r \<up> 2]",
     "SideConditions [(u::real) / 2 = r * sin \<alpha>, 2 / v = r * cos \<alpha>]",
   (*MethodC model:*)
     "FunctionVariable a", "FunctionVariable b", "FunctionVariable \<alpha>",
     "Domain {0 <..< r}",
     "Domain {0 <..< r}",
     "Domain {0 <..< \<pi> / 2}",
     "ErrorBound (\<epsilon> = (0::real))"]: TermC.as_string list,
    ("Diff_App", ["univariate_calculus", "Optimisation"], 
      ["Optimisation", "by_univariate_calculus"]): References.T)): Example.T
\<close> 
setup \<open>Know_Store.add_expls [(demo_example, ["Diff_App-No.123a"])]\<close>
ML \<open>
case Store.get (Know_Store.get_expls @{theory}) ["Diff_App-No.123a"] ["Diff_App-No.123a"] of
  (_, ("Constants [r = (7::real)]" :: "Maximum A" :: "AdditionalValues [u, v]" :: _, 
    ("Diff_App", ["univariate_calculus", "Optimisation"] ,["Optimisation", "by_univariate_calculus"]))) 
    => ()
| _ => error "intermed. example_store CHANGED";
\<close> ML \<open>
\<close> ML \<open>
(*
  Implementation reasonable, as soo as
 * clarified, how "empty" items in Model are represented
 * parse_references_input works
 * switching \<Odot>\<Otimes> i_model for probl and meth works.
*)
\<close> ML \<open>
val check_inputXXX = I_Model.complete (*in order *)
\<close> ML \<open>
fun check_input (*ctxt o-model ??*) (_: Problem.model_input_pos) = []: I_Model.T (*TODO*)
\<close> ML \<open>
\<close> ML \<open>
(* at the first encounter of an Example create the following data and transfer them to ctree:
    origin: eventually used by sub-problems, too
    O_Model: for check of missing items
    I_Model: for efficient check of user input
*)
\<close> ML \<open>
fun init_ctree example_id =
    let
      val thy = ThyC.get_theory_PIDE @{context} "Diff_App"
      val example_id' = References_Def.explode_id example_id
      val (_, (fmz as (model, refs as (_(*thy_id*), probl_id, _)))) =
        Store.get (Know_Store.get_expls @{theory}) example_id' example_id' 
      val ctxt = ContextC.initialise' thy model;

      val o_model = Problem.from_store ctxt probl_id |> #model |> O_Model.init thy model
    in
      Ctree.Nd (Ctree.PblObj {
        fmz = fmz, origin = (o_model, refs, TermC.empty (*TODO handle CAS_Cmd*)),
        spec = References.empty, probl = I_Model.empty, meth = I_Model.empty,
        ctxt = ctxt, loc = (NONE, NONE), 
        branch = Ctree.NoBranch, ostate = Ctree.Incomplete, result = (TermC.empty, [])}, [])
    end
\<close> ML \<open>
(*val update_state: string * (Problem.model_input_pos * References.T) ->
    Ctree.ctree -> Ctree.ctree*)
fun update_state (example_id, (model, refs)) Ctree.EmptyPtree = 
    init_ctree example_id |>  update_state (example_id, (model, refs))
  | update_state (_, (model, refs_input))
    (Ctree.Nd (Ctree.PblObj {fmz, origin = origin as (o_model, refs_orig, _),
      spec, probl, meth, ctxt, loc, branch, ostate, result}, children)) =
      let 
        val (thy_id, pbl_id, met_id) = References.select_from_input refs_input refs_orig spec
        val _ = "thy? re-parse all i_model ELSE do nothing"
        val _ = "pbl? switch from meth? check_input of probll ELSE do nothing"
        val _ = "met? switch from probl check_input of methl ELSE do nothing"
        val {model, ...} = Problem.from_store ctxt pbl_id;
        val complete_model = I_Model.complete o_model probl meth model
      in
        Ctree.Nd (Ctree.PblObj {
          fmz = fmz, origin = origin, spec = (thy_id, pbl_id, met_id), 
          probl = complete_model, meth = meth, (*TODO P_Model.switch_pbl_met*)
          ctxt = ctxt, loc = loc, 
          branch = branch, ostate = ostate, result = result}, children)
      end
  | update_state _ _ = raise ERROR "update_state: uncovered case in fun.def"
;
update_state: string * (Problem.model_input_pos * References.input) -> Ctree.ctree -> Ctree.ctree
\<close> ML \<open>
(* the first indication for a user model *)
datatype user_model_spec = Model_Refs | Model | None
val user_model_spec = Model_Refs
\<close> ML \<open>
fun show_template user_model_spec ((switch_pbl_met, p_model), input_refs) =
  case user_model_spec of 
    Model_Refs => "show Specification with Model and References"
      (* see example in VSCode_Example, subsubsection \<open>Complete Specification\<close>*)
  | Model => "show Specification with Model only, References"
       (* see example in VSCode_Example, \<open>Show the Model only, an intermediate mode\<close>*)
 | None => "show Specification with Model and References"
      (* see example in VSCode_Example, subsubsection \<open>Immediate Start of Interactive Solving\<close>*)
;
show_template: user_model_spec ->
    (P_Model.switch_pbl_met * P_Model.model_out) * References.input -> string
\<close> ML \<open>
val _ =
  Outer_Syntax.command \<^command_keyword>\<open>Example\<close>
    "prepare ISAC problem type and register it to Knowledge Store"
    (Parse.name -- Parse.!!! ( \<^keyword>\<open>Specification\<close> -- \<^keyword>\<open>:\<close> --
         \<^keyword>\<open>Model\<close> -- \<^keyword>\<open>:\<close>  |-- Problem.parse_pos_model_input --
           (\<^keyword>\<open>References\<close> -- \<^keyword>\<open>:\<close> |-- Parse.!!! References.parse_input_pos
             (** )-- ( \<^keyword>\<open>Solution\<close> -- keyword>\<open>:\<close>)( ** )WHAT IS WRONG HEREWITH?( **)) ) >>
      (fn (example_id, (model_input,
          ((thy_id, thy_id_pos), ((probl_id, probl_id_pos), (meth_id, meth_id_pos))))) =>
        Toplevel.theory (fn thy =>
          let
            val store = update_state (example_id, (model_input, (thy_id, probl_id, meth_id)))
                Ctree.EmptyPtree;
            (*---vvvvvvvvvvvvvv------- this comes from input \<Otimes> \<Odot> to Example *)
            val (switch_pbl_met, user_model_spec) = (P_Model.Model_For_Problem_Ref, Model_Refs)
            val args = P_Model.template_args switch_pbl_met (thy_id, probl_id, meth_id) store
            val _ = show_template user_model_spec args
          in set_data store thy end)));
\<close> ML \<open>
(*/-------------------------------------------------------------------------------------------\*)
Parse.name -- Parse.!!! ( \<^keyword>\<open>Specification\<close> -- \<^keyword>\<open>:\<close> --
         \<^keyword>\<open>Model\<close> -- \<^keyword>\<open>:\<close>  |-- Problem.parse_pos_model_input --
                  (\<^keyword>\<open>References\<close> -- \<^keyword>\<open>:\<close> |-- Parse.!!! References.parse_input_pos))
(*\-------------------------------------------------------------------------------------------/*)
(*/-------------------------------------------------------------------------------------------\*)
: Token.T list ->
    (string * (Problem.model_input_pos *
      ((string * Position.T) * ((string * Position.T) * (string * Position.T))))) * Token.T list
(*\-------------------------------------------------------------------------------------------/*) 
(*/-------------------------------------------------------------------------------------------\* )
   the type above is the type of the argument of \<open>(fn (_(*name*),..\<close> 
( *\------------------------------------------------------------------------------------------/*)
\<close> ML \<open>
\<close>


end
