(* Title: src/../gcd_poly_old.sml
   Author: Diana Meindl
   Copyright (c) Diana Meindl 2011

Programcode according to [1] for later lookup for mathematicians.
The tests below remain according to [1], 
while "$ISABELLE_ISAC_TEST/Tools/isac/Knowledge/gcd_poly_ml.sml" start exactly from the same state
and in time follows development in "$ISABELLE_ISAC/Knowledge/GCD_Poly_ML.thy".

This file includes *** tests ***.

[1] Franz Winkler. Polynomial Algorithms in Computer Algebra.
Springer-Verlag/Wien 1996.
*)

(*fun nth _ []      = error "nth _ []" (*Isabelle2002, still saved the hours of update*)
    | nth 1 (x::_) = x
    | nth n (_::xs) = nth (n- 1) xs;*)
  fun nth xs i = List.nth (xs, i);
"--------------------------------------------------------";
"table of contents --------------------------------------";
"--------------------------------------------------------";
"----------- auxiliary functions univariate -------------";
"=========== code for [1] p.93: univariate ==============";
"----------- auxiliary functions multivariate -----------";
"=========== code for [1] p.95: multivariate ============";
"--------------------------------------------------------";
"--- begin of univariate polynomials --------------------";
"----------- fun mod_inv --------------------------------";
"----------- fun CRA_2 ----------------------------------";
"----------- fun lc -------------------------------------";
"----------- fun deg ------------------------------------";
"----------- fun primeList ------------------------------";
"----------- fun find_next_prime_not_divide -------------";
"----------- fun poly_mod -------------------------------";
"----------- fun %* -------------------------------------";
"----------- fun %/ -------------------------------------";
"----------- fun %-% ------------------------------------";
"----------- fun %+% ------------------------------------";
"----------- fun CRA_2_poly -----------------------------";
"----------- fun mod_div --------------------------------";
"----------- fun lc_of_unipoly_not_0 --------------------";
"----------- fun %|% ------------------------------------";
"----------- fun mod_poly_gcd ---------------------------";
"----------- fun normirt_polynom ------------------------";
"----------- fun poly_centr -----------------------------";
"----------- fun mod_gcd_p ------------------------------";
"----------- fun gcd_n ----------------------------------";
"----------- fun pp -------------------------------------";
"----------- fun GCD_MOD---------------------------------";
"--- begin of multivariate polynomials ------------------";
"----------- fun get_coef --------------------------- ---";
"----------- fun get_varlist ----------------------------";
"----------- fun get_coeflist ---------------------------";
"----------- fun add_var_to_multipoly -------------------";
"----------- fun cero_multipoly -------------------------";
"----------- fun short_mv -------------------------------";
"----------- fun greater_var ----------------------------";
"----------- fun order_multipoly ------------------------";
"----------- fun lc_multipoly ---------------------------";
"----------- fun deg_multipoly --------------------------";
"----------- fun max_deg_var ----------------------------";
"----------- infix %%/ ----------------------------------";
"----------- fun cont -----------------------------------";
"----------- fun cont_multipoly -------------------------";
"----------- fun evaluate_mv ----------------------------";
"----------- fun mutli_to_uni ---------------------------";
"----------- fun find_new_r  ----------------------------";
"----------- fun newton_mv  -----------------------------";
"----------- fun mult_with_new_var  ---------------------";
"----------- fun greater_deg_multipoly  -----------------";
"----------- infix %%+%%  -------------------------------";
"----------- infix %%-%%  -------------------------------";
"----------- infix %%*%%  -------------------------------";
"----------- fun newton_first  --------------------------";
"----------- fun newton_mv  -----------------------------";
"----------- fun listgreater  ---------------------------";
"----------- finfix  %%|%%  -----------------------------";
"----------- fun GCD_MODm  ------------------------------";

(*text {*
  Below we reference 
  F. Winkler, Polynomial Algorithms. ...
  by page 93 and 95. The identifiers used in this book are re-used in this thesis
  in order to support reference (although some of these identifiers
  doe not conform with the Isabelle coding standards)
*}*)

 (*default_print_depth 3; 20*)
 type unipoly = int list;

"----------- auxiliary functions univariate -------------";
"----------- auxiliary functions univariate -------------";
"----------- auxiliary functions univariate -------------";

(*subsection {* calculations for integers *}*)

  infix div'
  fun a div' b = 
    if a < 0 then abs a div b * ~1
    else a div b;

(*subsection {* modulo calculations for integers *}*)
  fun mod_inv r m =
    let
      fun modi (r, rold, m, rinv) = 
        if rinv < m
        then
          if r mod m = 1
          then rinv
          else modi (rold * (rinv + 1), rold, m, rinv + 1)
        else 0
     in modi (r, r, m, 1) end;

  fun mod_div a b m =
    a * (mod_inv b m) mod m;

  fun aprox_root a =
    let fun root' a n = 
      if n*n < a then root' a (n+1) else n
    in root' a 1
  end

  (*  r = r1 mod m1 and r= r2 mod m2 *)
  fun CRA_2 (r1: int, m1: int, r2: int, m2: int ) =
   (r1 mod m1) + ((r2 - (r1 mod m1)) * (mod_inv m1 m2) mod m2) * m1

(*subsection {* prime opertations *}*)

  fun is_prime primelist number = 
    if length primelist >0
    then 
      if (number mod (nth primelist 0))=0
      then false
      else is_prime (nth_drop 0 primelist) number
    else true

  fun primeList number =
    let 
    fun make_primelist list last number =
      if (nth list (length list - 1)) < number
      then
        if ( is_prime list (last + 2)) 
        then make_primelist (list @ [(last + 2)]) (last + 2) number
        else make_primelist  list (last + 2) number
      else list
    in
      if number < 3
      then [2]
      else make_primelist [2,3] 3 number end
  
  (*find a prime greater p not dividing the number a*)
  fun find_next_prime_not_divide a p  = 
    let
      val next = nth (primeList (p + 1)) (length (primeList (p + 1)) - 1) ;
    in
      if a mod next  <> 0
        then next
      else find_next_prime_not_divide a next
  end

(*subsection {* calculations for univariate polynomials *}*)

  fun lc (uvp: unipoly) =
    if nth uvp (length uvp- 1) <>0
    then nth uvp (length uvp- 1)
    else lc (nth_drop (length uvp- 1) uvp);
  
  fun deg (uvp: unipoly) = 
    if nth uvp (length uvp- 1) <>0
    then length uvp - 1
    else deg (nth_drop (length uvp- 1) uvp)

  fun lc_of_unipoly_not_0 [] = [](* and delete lc=0*)
    | lc_of_unipoly_not_0 (uvp: unipoly) = 
      if nth uvp (length uvp - 1) =0
      then lc_of_unipoly_not_0 (nth_drop (length uvp- 1) uvp)
      else  uvp;

  fun normirt_polynom (poly1: unipoly) (m: int) =
    let
      val poly1 = lc_of_unipoly_not_0 poly1
      val lc_a=lc poly1;
      fun normirt poly1 b m lc_a i =
        if i=0
        then [mod_div (nth poly1 i) lc_a m]@b
        else normirt poly1 ( [mod_div(nth poly1 i) lc_a m]@b) m lc_a (i- 1) ;
    in 
      if length(poly1)=0
      then poly1
      else normirt poly1  [] m lc_a (length(poly1)- 1)
  end

   infix %*
  fun (a: unipoly) %* b =  map2 Integer.mult a (replicate (length (a)) b)
  
  infix %/ 
  fun (poly: unipoly) %/ b = (* =quotient*)
    let fun division poly b = poly div' b;
    in map2 division poly (replicate (length (poly)) b) end

  infix %-
  fun (poly: unipoly) %- b =
    let fun minus poly b = poly - b;
  in map2 minus poly (replicate (length (poly)) b) end

  infix %+%
  fun (a: unipoly) %+% (b: unipoly) =  map2 Integer.add a b

  infix %-%
  fun (a: unipoly) %-% (b: unipoly) =
    let fun minus a b = a-b;
    in  map2 minus a b end

  (* if poly2|poly1 *)
  infix %|%
  fun [b: int] %|% [a: int] = 
    if abs b<= abs a andalso a mod b = 0
    then true
    else false
  | (poly2: unipoly) %|% (poly1: unipoly) = 
    let 
    val b = (replicate (length poly1 - length(lc_of_unipoly_not_0 poly2)) 0) @ lc_of_unipoly_not_0 poly2 ;
    val c = lc poly1 div' lc b;
    val rest = lc_of_unipoly_not_0 (poly1 %-% (b %* c));
    in 
      if rest = []
      then true
      else
        if c<>0 andalso length rest >= length poly2
        then poly2 %|% rest 
        else false
    end

  fun poly_centr (poly: unipoly) (m: int) =
    let
      fun midle m =
        let val mid = m div' 2
        in 
          if m mod mid = 0
          then  mid
          else  mid+1
        end
      fun centr a m mid =
        if a > mid
        then a - m
        else a
      fun polyCentr poly poly' m mid counter =
        if length(poly) > counter
        then polyCentr poly (poly' @ [centr (nth poly counter) m mid]) m mid (counter+1)
        else (poly': unipoly)
    in polyCentr poly [] m (midle m) 0
  end

  fun sum_lmb (a: unipoly) exp =
    let fun  sum' [a] _ = a
           | sum' (a: int list) exp = 
      sum' ([((nth a 0)) + ( Integer.pow exp (nth a 1))] @ nth_drop 0 (nth_drop 0 a)) exp
    in sum' ([Integer.pow exp (nth a 0)] @ nth_drop 0 a) exp
  end;
Integer.min ;
  fun landau_mignotte_bound a b = 
  (Integer.pow (Integer.min (deg a) (deg b)) 2) * (abs (Integer.gcd (lc a) (lc b)))* 
  (Int.min (abs((aprox_root (sum_lmb a 2)) div ~(lc a)), abs(( aprox_root (sum_lmb b 2)) div ~(lc b))));

(*subsection {* modulo calculations for polynomials *}*)

  fun CRA_2_poly (ma, mb) (r1, r2) = 
  let 
    fun CRA_2' (ma, mb) (r1, r2) = CRA_2 (r1, ma,r2, mb)
  in  map (CRA_2' (ma, mb)) (r1 ~~ r2) end

  infix poly_mod
  fun uvp poly_mod m =
    let fun poly_mod' uvp m n =
      if n < length (uvp)
      then poly_mod' ((nth_drop 0 uvp)@[(nth uvp 0) mod m]) m ( n + 1 )
      else uvp (*end of poly_mod'*)
    in
      poly_mod' uvp m 0 end 

  fun mod_poly_gcd (upoly1: unipoly) (upoly2: unipoly) (m: int) =
    let 
      val moda = upoly1 poly_mod m;
      val modb = (replicate (length upoly1 - length(lc_of_unipoly_not_0 (upoly2 poly_mod m))) 0)
                  @ (lc_of_unipoly_not_0 (upoly2 poly_mod m)) ;
      val c =  mod_div (lc moda) (lc modb) m
      val rest = lc_of_unipoly_not_0 ((moda %-% (modb %*  c)) poly_mod m)
    in if rest = [] 
        then [upoly2, [0]]
        else
          if length rest < length upoly2
          then mod_poly_gcd upoly2 rest m 
          else mod_poly_gcd rest upoly2 m
    end

  fun mod_gcd_p (poly1: unipoly) (poly2: unipoly) (p: int) =
    let val gcd_p = mod_poly_gcd poly1 poly2 p
    in normirt_polynom (nth gcd_p 0) p 
    end

  fun gcd_n (up: unipoly) =
    if length up = 2
    then abs (Integer.gcd (nth up 0)(nth up 1))
    else gcd_n ([abs (Integer.gcd (nth up 0)(nth up 1))]@(nth_drop 0 (nth_drop 0 up)))

fun gcd_n up = abs (Integer.gcds up);

  fun pp [_: int] = [1]
    | pp (poly1: unipoly) = 
    if (poly1 poly_mod (gcd_n poly1)) = (replicate (length (poly1)) 0)
    then  poly1 %/ (gcd_n poly1)
    else poly1;

"=========== code for [1] p.93: univariate ==============";
"=========== code for [1] p.93: univariate ==============";
"=========== code for [1] p.93: univariate ==============";
(*subsection {* GCD_MOD Algorgithmus, code for [1] p.93: univariate *}*)

 fun GCD_MOD (a: unipoly) (b: unipoly) =
   if a = [0] orelse  b = [0] then [1] else
   if a = b then a else
   let
(*1*)val d = abs (Integer.gcd (lc a) (lc b));
     val M  = 2*d*landau_mignotte_bound a b;
    fun GOTO2 a b d M p =   (*==============================*)
       let 
(*2*)    val p = find_next_prime_not_divide d p
         val c_p = normirt_polynom ( mod_gcd_p a b p) p
         val g_p = poly_centr((c_p %* (d mod p)) poly_mod p) p
         fun GOTO3 a b d M p g_p =  (*~~~~~~~~~~~~~~~~~~~~~~*)
(*3*)      if (deg g_p) = 0
           then  [1]
           else
             let
               val P = p
               val g = g_p
               fun WHILE a b d M P g p = (*------------------*)
(*4*)            if P > M 
                 then g
                 else
                   let 
                     val p = find_next_prime_not_divide d p
                     val c_p = normirt_polynom ( mod_gcd_p a b p) p
                     val g_p = poly_centr((c_p %* (d mod p)) poly_mod p) p
                   in
                     if deg g_p < deg g
                     then GOTO3 a b d M p g_p
                     else
                       if deg g_p = deg g
                       then 
                         let 
                           val g = CRA_2_poly (P,p)(g,g_p)
                           val P = P*p
                           val g = poly_centr(g poly_mod P)P
                         in WHILE a b d M P g p end
                       else WHILE a b d M P g p 
               end (*----------------------------------*)
              
               
               val g = WHILE a b d M P g p (* << 1.Mal -----*)
             in g end (*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)
         
         val g = GOTO3 a b d M p g_p (* << 1.Mal ~~~~~~~~~~*)
(*5*)    val g = pp g
        in
          if (g %|% a) andalso (g %|% b)
          then  g
          else GOTO2 a b d M p
        end (*==============================================*)
      

     val  g = GOTO2 a b d  M (*p*)1(* << 1. Mal  =============*)
   in g end;

"----------- auxiliary functions multivariate -----------";
"----------- auxiliary functions multivariate -----------";
"----------- auxiliary functions multivariate -----------";
(*subsection {* GCD_MODm Algorgithmus, auxiliary functions multivariate *}*)

type monom = (int * int list);
type multipoly = monom list;

(*subsection {* calculations for multivariate polynomials *}*)

  fun listgreater a b = 
    let fun gr a b = a>=b;
    fun all [] = true 
      | all list = 
      if nth list 0 then all (nth_drop 0 list)  else false 
    in all (map2 gr a b)
  end
  
  fun get_coef (mvp: multipoly) (place: int) =
    let fun coef ((coef,_): monom) = coef;
    in coef (nth mvp place) end
  
  fun get_varlist (mvp: multipoly) (place: int) =
    let fun varlist ((_,var): monom) = var;
    in varlist (nth mvp place) end
  
  fun get_coeflist (poly: multipoly) = 
    let 
      fun get_coeflist' (poly: multipoly) list = 
        if poly = []
         then 
          list
         else          
           get_coeflist' (nth_drop 0 poly) (list @ [get_coef poly 0])
    in 
      get_coeflist' poly []
  end

  fun add_var_to_multipoly (mvp: multipoly) (order: int) =
    let fun add ([]: multipoly) (_: int) (new: multipoly) = new
          | add (mvp: multipoly) (order: int) (new: multipoly) =
          let val (first, last) = chop order (get_varlist mvp 0)
          in add (nth_drop 0 mvp) (order) (new @ [((get_coef mvp 0),(first @ [0] @ last))]) end 
    in add mvp order [] end

  fun cero_multipoly 1 = [(0,[0])]
    | cero_multipoly (diferent_var: int) = 
      add_var_to_multipoly (cero_multipoly (diferent_var- 1)) 0;

  fun short_mv (mvp: multipoly) = 
    let fun short (mvp: multipoly) (new: multipoly) =
              if length mvp =1 
              then if (get_coef mvp 0) = 0 
                   then if new = [] then cero_multipoly (length (get_varlist mvp 0)) else new
                   else  new @ mvp 
              else if get_varlist  mvp 0 = get_varlist mvp 1
                then short ( [(get_coef mvp 0 + get_coef mvp 1,get_varlist  mvp 0)]
                             @ nth_drop 0 (nth_drop 0 mvp))   new 
                else if (get_coef mvp 0) = 0 then short (nth_drop 0 mvp) new 
                  else short (nth_drop 0 mvp) (new @ [nth mvp 0])
    in short mvp [] end

  (* if a is greater than b *)
  fun greater_var [a: int] [b: int] = 
    a > b 
    | greater_var (a: int list) (b: int list) =
    if  (nth a (length a - 1))= (nth b (length b - 1))
      then greater_var (nth_drop (length a- 1) a) (nth_drop (length b- 1) b)
      else (nth a (length a - 1)) > (nth b (length b - 1))

  fun order_multipoly (a: multipoly)=
    let 
      val ordered = [nth a 0];
      fun order_mp [] [] = cero_multipoly (length (get_varlist a 0))
        | order_mp [] (ordered: multipoly) = short_mv ordered
        | order_mp a (ordered: multipoly) =
        if  greater_var (get_varlist a 0) (get_varlist ordered (length ordered - 1))
          then order_mp (nth_drop 0 a)(ordered @ [nth a 0])
          else let 
            val rest = [nth ordered (length ordered - 1)];
            fun order_mp' [] (new: multipoly) (rest: multipoly) = new @ rest
              | order_mp' (ordered': multipoly) (new: multipoly) (rest: multipoly) =
                if greater_var (get_varlist new 0) (get_varlist ordered' (length ordered' - 1))
                  then ordered' @ new @ rest
                  else order_mp' (nth_drop (length ordered' - 1) ordered') new 
                                 ([nth ordered' (length ordered' - 1)]@ rest)
            in order_mp (nth_drop 0 a) 
                        (order_mp' (nth_drop (length ordered - 1) ordered) [nth a 0] rest)
          end
    in order_mp (nth_drop 0 a) ordered
    end

  fun lc_multipoly (mpoly: multipoly) = 
   get_coef (order_multipoly mpoly) (length (order_multipoly mpoly) - 1);

 (* greatest variablegroup  *)
  fun deg_multipoly (mvp: multipoly) = 
    get_varlist (order_multipoly mvp) (length (order_multipoly mvp) - 1)
 
  fun  max_deg_var [m: monom] (x: int) = nth (get_varlist [m] 0) x |
   max_deg_var (mvp: multipoly) (x: int) =
    let
      fun max_monom (m1: monom) (m2: monom) (x: int)=
        if nth (get_varlist [m1] 0) x < nth (get_varlist [m2] 0) x  then m2  else m1;
    in
    max_deg_var ([max_monom (nth(mvp) 0)( nth(mvp) 1) x] @ nth_drop 0 (nth_drop 0 mvp)) x
    end  

  fun greater_deg (monom1: monom) (monom2: monom)=
    if (max_deg_var [monom1] (length(get_varlist [monom1] 0) - 1)) > 
       (max_deg_var [monom2] (length (get_varlist [monom2] 0) - 1))
      then 1
      else if (max_deg_var [monom1] (length(get_varlist [monom1] 0) - 1)) < 
              (max_deg_var [monom2] (length (get_varlist [monom2] 0) - 1))
        then 2
        else if length (get_varlist [monom1] 0) >1
          then
            greater_deg (1,(nth_drop 0 (get_varlist [monom1] 0))) (1,(nth_drop 0 (get_varlist [monom2] 0)))
          else 0
          
  infix %%/
  fun (poly: multipoly) %%/ b = (* =quotient*)
    let fun division monom b = (((get_coef [monom] 0) div' b),get_varlist [monom] 0);
    in order_multipoly(map2 division poly (replicate (length(poly)) b)) end

  fun cont [a: int] = a
    | cont (poly1: unipoly) = gcd_n poly1
  fun cont_multipoly [(a,_): monom] = a 
    | cont_multipoly (poly: multipoly) = gcd_n (get_coeflist poly)

  fun evaluate_mv (mvp: multipoly) (var: int) (value: int) = 
    let fun eval ([]: multipoly) (_: int) (_: int) (new: multipoly) = order_multipoly new |
      eval (mvp: multipoly) (var: int) (value: int) (new: multipoly) =
      eval (nth_drop 0 mvp) var value
           (new @  [((Integer.pow  (nth (get_varlist mvp 0) var)value) * get_coef mvp 0, 
                    nth_drop var (get_varlist mvp 0))]);
    in eval mvp var value [] end

  fun multi_to_uni (mvp: multipoly) = 
    if length (get_varlist mvp 0) = 1 
    then let fun mtu ([]: multipoly) (uvp: unipoly) = uvp |
                 mtu (mvp: multipoly) (uvp: unipoly) =
               if length uvp = (nth(get_varlist mvp 0) 0)
               then mtu (nth_drop 0 mvp) (uvp @ [get_coef mvp 0])
               else mtu mvp (uvp @ [0])    
         in mtu (order_multipoly mvp) [] end
    else error "Polynom has more than one variable!";

  fun uni_to_multi (uvp: unipoly) =
    let fun utm ([]: unipoly) (mvp: multipoly) (_: int)=  short_mv mvp
          | utm (uvp: unipoly) (mvp: multipoly) (counter: int) = 
          utm (nth_drop 0 uvp) (mvp @ [((nth uvp 0),[counter])]) (counter+1)
    in  utm uvp [] 0 end
 
  fun find_new_r (mvp1: multipoly) (mvp2: multipoly) (old: int) =
    let val poly1 = evaluate_mv mvp1 (length (get_varlist mvp1 0) - 2) (old + 1)
        val poly2 = evaluate_mv mvp2 (length (get_varlist mvp2 0) - 2) (old + 1);
    in
      if max_deg_var poly1 (length (get_varlist poly1 0) - 1)= max_deg_var mvp1 (length (get_varlist mvp1 0) - 1) orelse
         max_deg_var poly2 (length (get_varlist poly2 0) - 1)= max_deg_var mvp2 (length (get_varlist mvp2 0) - 1) 
      then old + 1
      else find_new_r (mvp1) (mvp2) (old + 1)
    end

  fun mult_with_new_var ([]: multipoly) (_: unipoly) (_: int) = []
    | mult_with_new_var (mvp: multipoly) (uvp: unipoly) (order: int) = 
    let fun mult ([]: multipoly) (_: unipoly) (_: int) (new: multipoly) (_: int) = short_mv new
          | mult (mvp: multipoly) (uvp: unipoly) (order: int) (new: multipoly) (_: int)  =
          let fun mult' (_: multipoly) ([]: unipoly) (_: int) (new': multipoly) (_) = order_multipoly new'
                | mult' (mvp': multipoly) (uvp': unipoly) (order: int) (new': multipoly) (c2': int) =
                let val (first, last) = chop order (get_varlist mvp' 0)
                in mult' mvp' (nth_drop 0 uvp') order
                        (new' @ [(((get_coef mvp' 0) * (nth uvp' 0)),(first @ [c2'] @ last))]) (c2'+1) end
          in mult (nth_drop 0 mvp) uvp order (new @ (mult' mvp uvp order [] 0)) (0) end
    in mult mvp uvp order [] 0 end
  
  fun greater_deg_multipoly (var1: int list) (var2: int list) =
    if var1 = [] andalso var2 =[] then 0
    else if (nth var1 (length var1 - 1)) = (nth var2 (length var1 - 1) ) 
         then greater_deg_multipoly (nth_drop (length var1 - 1) var1) (nth_drop (length var1 - 1) var2)
         else if (nth var1 (length var1 - 1)) > (nth var2 (length var1 - 1)) 
              then 1 else 2 ;

  infix %%+%%
  fun ([]: multipoly) %%+%% (mvp2: multipoly) = mvp2
    | (mvp1: multipoly) %%+%% ([]: multipoly) = mvp1
    | (mvp1: multipoly) %%+%% (mvp2: multipoly) =
    let fun plus ([]: multipoly) (mvp2: multipoly) (new: multipoly) = order_multipoly (new @ mvp2)
          | plus (mvp1: multipoly) ([]: multipoly) (new: multipoly) = order_multipoly (new @ mvp1)
          | plus (mvp1: multipoly) (mvp2: multipoly) (new: multipoly) = 
          if greater_deg_multipoly (get_varlist mvp1 0) (get_varlist mvp2 0) = 0
          then plus (nth_drop 0 mvp1) (nth_drop 0 mvp2) 
                    (new @ [(((get_coef mvp1 0) + (get_coef mvp2 0)), get_varlist mvp1 0)])
          else if greater_deg_multipoly (get_varlist mvp1 0) (get_varlist mvp2 0) = 1
               then plus mvp1 (nth_drop 0 mvp2) (new @ [nth mvp2 0])
               else plus (nth_drop 0 mvp1) mvp2 (new @ [nth mvp1 0])
    in plus mvp1 mvp2 [] end
  
  infix %%-%%
  fun (mvp1: multipoly) %%-%% (mvp2: multipoly) = 
    let fun neg ([]: multipoly) (new: multipoly) = order_multipoly new
          | neg (mvp: multipoly) (new: multipoly) =
          neg (nth_drop 0 mvp) (new @ [(((get_coef mvp 0) * ~1), get_varlist mvp 0)])
        val neg_mvp2 = neg mvp2
    in mvp1 %%+%% (neg_mvp2 [])  end         
          
  infix %%*%%
  fun (mvp1: multipoly) %%*%% (mvp2: multipoly) =
    let fun mult ([]: multipoly) (_: multipoly) (_: multipoly) (new: multipoly) = order_multipoly new
          | mult (mvp1: multipoly) ([]: multipoly) (regular_mvp2: multipoly) (new: multipoly) = mult (nth_drop 0 mvp1) regular_mvp2 regular_mvp2 new
          | mult (mvp1: multipoly) (mvp2: multipoly) (regular_mvp2: multipoly) (new: multipoly) = 
          mult mvp1 (nth_drop 0 mvp2) regular_mvp2 (new @ [(((get_coef mvp1 0) * (get_coef mvp2 0)), ((get_varlist mvp1 0) %+% (get_varlist mvp2 0)))])
  in if (length mvp1) > (length mvp2) then mult mvp1 mvp2 mvp2 [] else mult mvp2 mvp1 mvp1 [] end

 (* if poly1|poly2 *)
  infix %%|%% 
  fun [(coef2, var2): monom] %%|%%  [(coef1, var1): monom]  = 
    ( (listgreater var1 var2) 
        andalso (coef1 mod coef2) = 0)
    | (_: multipoly) %%|%% [(0,_)]=  true
    | (poly2: multipoly)  %%|%% (poly1: multipoly) =
    if [nth poly2 (length poly2 - 1)]   %%|%% [nth poly1 (length poly1 - 1)]
    then poly2 %%|%% (poly1 %%-%%
      ([((get_coef poly1 (length poly1 - 1)) div (get_coef poly2 (length poly2 - 1)),
      (get_varlist poly1 (length poly1 - 1)) %-%(get_varlist poly2 (length poly2 - 1)))] %%*%%
      poly2)) 
    else false

(*subsection {* Newtoninterpolation *}*)
 (* first step *)
  fun newton_first (x: int list) (f: multipoly list) (order: int) =
    let val polynom =(add_var_to_multipoly (nth f 0) order) %%+%% 
            ((mult_with_new_var (((nth f 1)%%-%% (nth f 0))%%/
               (nth x 1) - (nth x 0))) [(nth x 0) * ~1, 1] order);
        val new_value_poly =   [(nth x 0) * ~1, 1];
        val steps = [((nth f 1) %%-%%(nth f 0))%%/ ((nth x 1) - (nth x 0))]
    in (polynom, new_value_poly, steps) end
  
  fun newton_mv (x: int list) (f: multipoly list) (steps: multipoly list) (t: unipoly) (p: multipoly) (order: int) = 
    if length x = 2
    then let val (polynom, new_value_poly, steps) =  newton_first x [(nth f 0), (nth f 1)] order
         in (polynom, new_value_poly, steps) end
    else let val new_value_poly = multi_to_uni((uni_to_multi t)  %%*%% (uni_to_multi [(nth x (length x - 2) )* ~1, 1]));
             val new_steps = [((nth f (length f - 1))  %%-%% (nth f (length f - 2)))  %%/ ((nth x (length x - 1)) - (nth x (length x - 2)))];
             fun next_step ([]: multipoly list) (new_steps: multipoly list) (_: int list) = new_steps
               | next_step (steps: multipoly list) (new_steps: multipoly list) (x': int list) =  
                 next_step (nth_drop 0 steps)
                           (new_steps @ [(((nth new_steps (length new_steps - 1)) %%-%%(nth steps 0)) %%/
                                    ((nth x' (length x' - 1)) - (nth x' (length x' - 3))))])
                           ( nth_drop (length x' - 2) x')
             val steps = next_step steps new_steps x;
             val polynom' =  (p %%+%% (mult_with_new_var (nth steps (length steps - 1)) new_value_poly order));
       in (order_multipoly(polynom'), new_value_poly, steps) end;

"=========== code for [1] p.95: multivariate ============";
"=========== code for [1] p.95: multivariate ============";
"=========== code for [1] p.95: multivariate ============";
(*subsection {* GCD_MODm algorithm, code for [1] p.95: multivariate *}*)

fun GCD_MODm a b n s r=
  if greater_var (deg_multipoly b) (deg_multipoly a) then GCD_MODm b a n s r else
(*0*)  if s = 0
    then  uni_to_multi((GCD_MOD (pp(multi_to_uni a)) (pp(multi_to_uni b))) %* 
          (abs (Integer.gcd (cont (multi_to_uni a)) (cont( multi_to_uni b)))))
      else 
      let 
(*1*)   val M = 1 + Int.min (max_deg_var a (length(get_varlist a 0)- 2), max_deg_var b (length(get_varlist b 0)- 2)); 
(*2*)     fun GOTO2 a b n s M r_list steps =  (*==============================*)
              let 
                val r = find_new_r a b r;
                val r_list = r_list @ [r];
                val g_r = GCD_MODm (order_multipoly (evaluate_mv a (s- 1) r)) ( order_multipoly (evaluate_mv b (s- 1) r)) (n- 1) (s- 1) 0
                val steps = steps @ [g_r];
(*3*)           fun GOTO3 a b n s M g_r r r_list steps = (* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
                  let  
                    val m = 1;
                    fun WHILE a b n s M m r r_list steps g g_n mult =   (* ----------------------- *)
                      if m > M
                        then 
                         if g_n %%|%% a andalso g_n %%|%% b
                          then  g_n
                          else GOTO2 a b n s M r_list steps
                        else
                        let 
                        val r = find_new_r a b r; 
                        val r_list = r_list @ [r];
                        val g_r = GCD_MODm (order_multipoly  (evaluate_mv a (s- 1) r))
                                            (order_multipoly  (evaluate_mv b (s- 1) r)) (n- 1) (s- 1) 0
                        in  
                          if greater_var  (deg_multipoly g) (deg_multipoly g_r)
                            then GOTO3 a b n s M g_r r r_list steps
                            else if (deg_multipoly g)= (deg_multipoly g_r)
                              then
                                let 
                                  val (g_n, new, steps) = newton_mv r_list [g, g_r] steps mult g_n (s- 1)
                                in if (nth steps (length steps - 1)) = (cero_multipoly s)
                                   then WHILE a b n s M (M+1) r r_list steps g_r g_n new
                                   else WHILE a b n s M (m+1) r r_list steps g_r g_n new 
                                end
                              else WHILE a b n s M (m+1) r r_list steps g  g_n mult
                        end (* WHILE *)
                  in (* GOTO3*) (* ----------------------- *)
                    WHILE a b n s M m r r_list steps g_r ( cero_multipoly (s+1)) [1]
                  end (*GOTO3*)
              in (* GOTO2*)  (* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
                GOTO3 a b n s M g_r r r_list steps
              end (*GOTO2*)                              
      in  (*==============================*)
      GOTO2 a b n s M [] []
    end (*end 0*);


(******************************************** tests ********************************************)
(******************************************** tests ********************************************)
(******************************************** tests ********************************************)

"----------- fun mod_inv --------------------------------";
"----------- fun mod_inv --------------------------------";
"----------- fun mod_inv--- -----------------------------";
"~~~~~ fun mod_inv, args:"; val (r, m) = (5,7);
"~~~~~ fun modi, args:"; val (r, rold, m, rinv) = (r, r, m, 1);
rinv < m; (*=true*)
r mod m = 1; (*=false*)
"~~~~~ fun modi, args:"; val (r, rold, m, rinv) = (rold * (rinv + 1), rold, m, rinv + 1);
rinv < m;(*=true*)
r mod m = 1; (*=false*)
"~~~~~ fun modi, args:"; val (r, rold, m, rinv) = (rold * (rinv + 1), rold, m, rinv + 1);
rinv < m;(*=true*)
r mod m = 1; (*=true*)
"~~~~~ to mod_inv return val:"; val (rinv) = (rinv);

if mod_inv 5 7 = 3 then () else error "mod_inv 5 7 = 3 changed";
if mod_inv 3 7 = 5 then () else error "mod_inv 3 7 = 5 changed";
if mod_inv 4 339 = 85 then () else error "mod_inv 4 339 = 85 changed";

"----------- fun CRA_2 ----------------------------------";
"----------- fun CRA_2 ----------------------------------";
"----------- fun CRA_2 ----------------------------------";

if  CRA_2(17,9,3,4) = 35 then () else error "CRA_2(17,9,3,4)=35 changed"; 
if  CRA_2(7,2,6,11) = 17  then () else error "CRA_2(7,2,6,11) = 17 changed";  

"----------- fun lc -------------------------------------";
"----------- fun lc -------------------------------------";
"----------- fun lc -------------------------------------";
if lc [3,4,5,6] = 6 then () else error "lc (3,4,5,6) = 6 changed" ;
if lc [3,4,5,6,0] = 6 then () else error "lc (3,4,5,6,0) = 6 changed"  ;

"----------- fun deg ------------------------------------";
"----------- fun deg ------------------------------------";
"----------- fun deg ------------------------------------";
if deg [3,4,5,6] = 3 then () else error "lc (3,4,5,6) = 3 changed" ;
if deg [3,4,5,6,0] = 3 then () else error "lc (3,4,5,6) = 6 changed";

"----------- fun primeList ------------------------------";
"----------- fun primeList ------------------------------";
"----------- fun primeList ------------------------------";
if primeList 1 = [2] then () else error " primeList 1 = [2] changed";
if primeList 3 = [2,3] then () else error " primeList 3 = [2,3] changed";
if primeList 6 = [2,3,5,7] then () else error " primeList 6 = [2,3,5,7] changed";
if primeList 7 = [2,3,5,7] then () else error " primeList 7 = [2,3,5,7] changed";

"----------- fun find_next_prime_not_divide -----------------";
"----------- fun find_next_prime_not_divide -----------------";
"----------- fun find_next_prime_not_divide -----------------";
if find_next_prime_not_divide 15 2 = 7 then () else error "findNextPrimeNotdivide 15 2 = 7 changed";
if find_next_prime_not_divide 90 1 = 7 then () else error "findNextPrimeNotdivide 90 1 = 7 changed";

"----------- fun poly_mod -------------------------------";
"----------- fun poly_mod -------------------------------";
"----------- fun poly_mod -------------------------------";
if ([5,4,7,8,1] poly_mod 5) = [0, 4, 2, 3, 1] then () 
else error "[5,4,7,8,1] poly_mod 5 = [0, 4, 2, 3, 1] changed" ;
if ([5,4,~7,8,~1] poly_mod 5) = [0, 4, 3, 3, 4] then () 
else error "[5,4,~7,8,~1] poly_mod 5  = [0, 4, 3, 3, 4] changed" ;

"----------- fun %* ------------------------------";
"----------- fun %* ------------------------------";
"----------- fun %* ------------------------------";
if ([5,4,7,8,1] %* 5) = [25, 20, 35, 40, 5] then () 
else error "[5,4,7,8,1] %* 5 = [25, 20, 35, 40, 5] changed";
if ([5,4,~7,8,~1] %* 5) = [25, 20, ~35, 40, ~5] then () 
else error "[5,4,~7,8,~1] %* 5 = [25, 20, ~35, 40, ~5] changed";

"----------- fun %/ -------------------------------";
"----------- fun %/ -------------------------------";
"----------- fun %/ -------------------------------";
if ([4,3,2,5,6] %/ 3) =[1, 1, 0, 1, 2] then ()
else error "%/ [4,3,2,5,6] 3 =[1, 1, 0, 1, 2] changed";
if ([4,3,2,0] %/ 3) =[1, 1, 0, 0] then () 
else error "%/ [4,3,2,0] 3 =[1, 1, 0, 0] changed";

"----------- fun %-% ------------------------";
"----------- fun %-% ------------------------";
"----------- fun %-% ------------------------";
if ([8,~7,0,1] %-% [~2,2,3,0])=[10,~9,~3,1] then () 
else error "[8,~7,0,1] %-% [~2,2,3,0]=[10,~9,~3,1] changed";
if ([8,7,6,5,4] %-% [2,2,3,1,1])=[6,5,3,4,3] then ()
else error "[8,7,6,5,4] %-% [2,2,3,1,1]=[6,5,3,4,3] changed";

"----------- fun %+% ------------------------------";
"----------- fun %+% ------------------------------";
"----------- fun %+% ------------------------------";
if ([8,~7,0,1] %+% [~2,2,3,0])=[6,~5,3,1] then ()
else error "[8,~7,0,1] %+% [~2,2,3,0]=[6,~5,3,1] changed";
if ([8,7,6,5,4] %+% [2,2,3,1,1])=[10,9,9,6,5] then ()
else error "[8,7,6,5,4] %+% [2,2,3,1,1]=[10,9,9,6,5] changed";

"----------- fun CRA_2_poly -----------------------------";
"----------- fun CRA_2_poly -----------------------------";
"----------- fun CRA_2_poly -----------------------------";
if (CRA_2_poly (5, 7)([2,2,4,3],[3,2,3,5]))=[17,2,24,33] then ()
else error "CRA_2_poly (5, 7)([2,2,4,3],[3,2,3,5])=[17,2,24,33] changed";

"----------- fun mod_div --------------------------------";
"----------- fun mod_div --------------------------------";
"----------- fun mod_div --------------------------------";
if mod_div 21 4 5 = 4 then () else error "mod_div 21 4 5 = 4 changed";
if mod_div 1 4 5 = 4 then () else error "mod_div 1 4 5 = 4 changed";
if mod_div 0 4 5 = 0 then () else error "mod_div 0 4 5 = 0 changed";

"----------- fun lc_of_unipoly_not_0 --------------------";
"----------- fun lc_of_unipoly_not_0 --------------------";
"----------- fun lc_of_unipoly_not_0 --------------------";
if lc_of_unipoly_not_0 [0,1,2,3,4,5,0,0]=[0,1, 2, 3, 4, 5] then ()
else error "lc_of_unipoly_not_0 [0,1,2,3,4,5,0,0]=[0,1, 2, 3, 4, 5] changed";
if lc_of_unipoly_not_0 [0,1,2,3,4,5]=[0,1, 2, 3, 4, 5] then () 
else error "lc_of_unipoly_not_0 [0,1,2,3,4,5]=[0, 1, 2, 3, 4, 5] changed";

"----------- fun %|% -------------------------------";
"----------- fun %|% -------------------------------";
"----------- fun %|% -------------------------------";
"~~~~~ fun %|% , args:"; val (poly1, poly2) = ([9,12,4],[3,2]);
val b = (replicate (length poly1 - length poly2) 0 ) @ poly2 ;
val c = lc poly1 div lc b;
val rest = lc_of_unipoly_not_0 ( poly1 %-% (b %* c ));
rest = [];(*=false*)
c<>0 andalso length rest >= length poly2; (*=true*)
"~~~~~ fun %|% , args:"; val (poly1, poly2) = (rest, poly2);
val b = (replicate (length poly1 - length poly2) 0 ) @ poly2 ;
val c = lc poly1 div lc b;
val rest = lc_of_unipoly_not_0 ( poly1 %-% (b %* c ));
rest = [];(*=true*)
"~~~~~ to  return val:"; val (divide) = (true);

if ([4] %|% [6]) = false then () else error "[4] %|% [6] = false changed";
if ( [8] %|%[16,0]) = true then () else error "[8] %|%[16,0] = true changed";
if ([3,2] %|%[0,0,9,12,4] ) = true then () 
else error "[3,2] %|%[0,0,9,12,4]  = true changed";
if ([8,0] %|% [16]) = true then () else error "[8,0] %|% [16] = true changed";

"----------- fun mod_poly_gcd ------------------------";
"----------- fun mod_poly_gcd ------------------------";
"----------- fun mod_poly_gcd ------------------------";

if ( mod_poly_gcd [~18, ~15, ~20, 12, 20, ~13, 2] [8, 28, 22, ~11, ~14, 1, 2] 7)=[[2, 6, 0, 2, 6], [0]]
then () else error "( mod_poly_gcd [~18, ~15, ~20, 12, 20, ~13, 2] [8, 28, 22, ~11, ~14, 1, 2] 7 = [ [5, 1, 0, 5, 1], [0]] changed";
if ( mod_poly_gcd [8, 28, 22, ~11, ~14, 1, 2] [2, 6, 0, 2, 6] 7)=[[2, 6, 0, 2, 6], [0]] then ()
else error "mod_poly_gcd [8, 28, 22, ~11, ~14, 1, 2] [2, 6, 0, 2, 6] 7=[[1, 3, 0, 1, 3], [0]] changed";
if mod_poly_gcd [20,15,8,6] [8,~2,~2,3] 2 = [[0, 1], [0]] then ()
else error " mod_poly_gcd [20,15,8,6] [8,~2,~2,3] 2 = [[0, 1], [0]] changed";

"~~~~~ fun mod_poly_gcd , args:"; 
val (a,b,m) = ([ ~13, 2,12],[ ~14, 2],5);
val moda = a poly_mod m;
val modb = (replicate (length a - length(lc_of_unipoly_not_0 b)) 0) @ (lc_of_unipoly_not_0 b poly_mod m) ;
val c =  mod_div (lc moda) (lc modb) m;
val rest = lc_of_unipoly_not_0 (moda %-% (modb %* c) poly_mod m) poly_mod m;
rest = []; (*=false*)
length rest < length b; (*=false*)
"~~~~~ fun  mod_poly_gcd , args:";
val (a,b,m,d) = (rest, b ,  m , [c]);
val moda = a poly_mod m
val modb = (replicate (length a - length(lc_of_unipoly_not_0 b)) 0) @ (lc_of_unipoly_not_0 b poly_mod m) ;
val c =  mod_div (lc moda) (lc modb) m
val rest = lc_of_unipoly_not_0 ((moda %-% (modb %*  c)) poly_mod m);
rest = [];(*=flase*)
length rest < length b; (*=true*)
"~~~~~ fun  mod_poly_gcd , args:";
val (a,b,m,d) = (b, rest, m, [c] @ d);
val moda = a poly_mod m
val modb = (replicate (length a - length(lc_of_unipoly_not_0 b)) 0) @ (lc_of_unipoly_not_0 b poly_mod m) ;
val c =  mod_div (lc moda) (lc modb) m
val rest = lc_of_unipoly_not_0 ((moda %-% (modb %*  c)) poly_mod m);
rest = [];(*=flase*)
length rest < length b; (*=false*)
"~~~~~ fun  mod_poly_gcd , args:";
val (a,b,m,d) = (b, rest, m, [c] @ d);
val moda = a poly_mod m
val modb = (replicate (length a - length(lc_of_unipoly_not_0 b)) 0) @ (lc_of_unipoly_not_0 b poly_mod m) ;
val c =  mod_div (lc moda) (lc modb) m
val rest = lc_of_unipoly_not_0 ((moda %-% (modb %*  c)) poly_mod m);
rest = [];(*=true*)
"~~~~~ to  return val:"; val ([b, rest, lsit])=[b, [0], [c] @ d];


"----------- fun normirt_polynom ------------------------";
"----------- fun normirt_polynom ------------------------";
"----------- fun normirt_polynom ------------------------";
if (normirt_polynom [~18, ~15, ~20, 12, 20, ~13, 2] 5) =[1, 0, 0, 1, 0, 1, 1 ] then ()
else error "(normirt_polynom [~18, ~15, ~20, 12, 20, ~13, 2] 5) =[1, 0, 0, 1, 0, 1, 1 ] changed"

"----------- fun poly_centr -----------------------------";
"----------- fun poly_centr -----------------------------";
"----------- fun poly_centr -----------------------------";
if (poly_centr [7,3,5,8,1,3] 10) = [~3, 3, 5, ~2, 1, 3] then ()
else error "poly_centr [7,3,5,8,1,3] 10 = [~3, 3, 5, ~2, 1, 3] cahnged";

"----------- fun mod_gcd_p -------------------------------";
"----------- fun mod_gcd_p -------------------------------";
"----------- fun mod_gcd_p -------------------------------";
if  (mod_gcd_p [~18, ~15, ~20, 12, 20, ~13, 2] [8, 28, 22, ~11, ~14, 1, 2] 5) = [1, 1, 1, 1] then ()
else error "mod_gcd_p [~18, ~15, ~20, 12, 20, ~13, 2] [8, 28, 22, ~11, ~14, 1, 2] 5 = [1, 1, 1, 1] changed";
if  (mod_gcd_p [~18, ~15, ~20, 12, 20, ~13, 2] [8, 28, 22, ~11, ~14, 1, 2] 7) = [5, 1, 0, 5, 1] then ()
else error "mod_gcd_p [~18, ~15, ~20, 12, 20, ~13, 2] [8, 28, 22, ~11, ~14, 1, 2] 7 = [5, 1, 0, 5, 1] changed";


"----------- fun gcd_n ----------------------------";
"----------- fun gcd_n ----------------------------";
"----------- fun gcd_n ----------------------------";
if gcd_n [3,9,12,21] = 3 then () else error "gcd_n [3,9,12,21] = 3 changed";
if gcd_n [12,16,32,44] = 4 then () else error "gcd_n [12,16,32,44] = 4 changed";
if gcd_n [4,5,12] = 1 then () else error "gcd_n [4,5,12] = 1 changed";

"----------- fun pp -------------------------------------";
"----------- fun pp -------------------------------------";
"----------- fun pp -------------------------------------";
if pp [12,16,32,44] = [3, 4, 8, 11] then () else error "pp [12,16,32,44] = [3, 4, 8, 11] changed";
if pp [4,5,12] =  [4, 5, 12] then () else error "pp [4,5,12] =  [4, 5, 12] changed";

"----------- fun GCD_MOD --------------------------------";
"----------- fun GCD_MOD --------------------------------";
"----------- fun GCD_MOD --------------------------------";

"~~~~~ fun GCD_MOD a b , args:"; val (a, b) = ([~18, ~15, ~20, 12, 20, ~13, 2], [8, 28, 22, ~11, ~14, 1, 2]);
val d = abs (Integer.gcd (lc a) (lc b));
val M  =2*d* landau_mignotte_bound a b;
(*val  g = GOTO2 a b d (*p*)1 M*)
   "~~~~~ fun GOTO2 a b d M p , args:"; val (a, b, d, p, M) = (a,b,d,3,M);
  val p = find_next_prime_not_divide d p
  val c_p = normirt_polynom ( mod_gcd_p a b p) p
  val g_p = poly_centr((c_p %* (d mod p)) poly_mod p) p;
  (*val (p, g) = GOTO3 a b d M p g_p (* << 1.Mal ~~~~~~~~~~*)*)       
     "~~~~~ fun GOTO3 a b d M p g_p , args:"; val (a, b, d, M, p, g_p) = (a,b,d,M,p,g_p);
    (deg g_p) = 0; (*=false*)
    val P = p
    val g = g_p;
    (*(p, g) = WHILE a b d M P g (* << 1.Mal -----*)*)
       "~~~~~ fun WHILE a b d M p g , args:"; val (a, b, d, M, P, g) = (a,b,d,M,p,g);
       P > M ;(*false*)
      val p = find_next_prime_not_divide d p
      val c_p = normirt_polynom ( mod_gcd_p a b p) p
      val g_p = poly_centr((c_p %* (d mod p)) poly_mod p) p;
      deg g_p < deg g;   (*=fasle*) 
      deg g_p = deg g; (*false*);
      "~~~~~ fun WHILE a b d M p g , args:"; val (a, b, d, M, P, g) = (a,b,d,M,P,g);
       P > M ;(*false*)
      val p = find_next_prime_not_divide d p
      val c_p = normirt_polynom ( mod_gcd_p a b p) p
      val g_p = poly_centr((c_p %* (d mod p)) poly_mod p) p;
      deg g_p < deg g;   (*=false*) 
      deg g_p = deg g; (*=true*)
      val g = CRA_2_poly (P,p)(g,g_p)
      val P = P*p;
      val g = poly_centr(g poly_mod P)P;
      "~~~~~ fun WHILE a b d M p g , args:"; val (a, b, d, M, P, g) = (a,b,d,M,P,g);
       P > M ;(*false*)
      val p = find_next_prime_not_divide d p
      val c_p = normirt_polynom (mod_gcd_p a b p) p
      val g_p = poly_centr((c_p %* (d mod p)) poly_mod p) p;
      deg g_p < deg g;   (*=true*) 
       "~~~~~ fun GOTO3 a b d M p g_p , args:"; val (a, b, d, M, p, g_p) = (a,b,d,M,p,g_p);
      (deg g_p) = 0; (*false*)
      val P = p
      val g = g_p;
      "~~~~~ fun WHILE a b d M p g , args:"; val (a, b, d, M, P, g) = (a,b,d,M,P,g);
       P > M ;(*false*)
      val p = find_next_prime_not_divide d p
      val c_p = normirt_polynom (mod_gcd_p a b p) p
      val g_p = poly_centr((c_p %* (d mod p)) poly_mod p) p;
      deg g_p < deg g;   (*=fasle*) 
      deg g_p = deg g;(*true*)
      val g = CRA_2_poly (P,p)(g,g_p)
      val P = P*p;
      val g = poly_centr(g poly_mod P)P;
      "~~~~~ fun WHILE a b d M p g , args:"; val (a, b, d, M, P, g) = (a,b,d,M,P,g);
       P > M ;(*false*)
      val p = find_next_prime_not_divide d p
      val c_p = normirt_polynom (mod_gcd_p a b p) p
      val g_p = poly_centr((c_p %* (d mod p)) poly_mod p) p;
      deg g_p < deg g;   (*=false*) 
      deg g_p = deg g;(*true*)
      val g = CRA_2_poly (P,p)(g,g_p)
      val P = P*p;
      val g = poly_centr(g poly_mod P)P;
      "~~~~~ fun WHILE a b d M p g , args:"; val (a, b, d, M, P, g) = (a,b,d,M,P,g);
       P > M ;(*false*)
      val p = find_next_prime_not_divide d p
      val c_p = normirt_polynom (mod_gcd_p a b p) p
      val g_p = poly_centr((c_p %* (d mod p)) poly_mod p) p;
      deg g_p < deg g;   (*=false*) 
      deg g_p = deg g;(*true*)
      val g = CRA_2_poly (P,p)(g,g_p)
      val P = P*p;
      val g = poly_centr(g poly_mod P)P;
      "~~~~~ fun WHILE a b d M p g , args:"; val (a, b, d, M, P, g) = (a,b,d,M,P,g);
       P > M ;(*true*)
    "~~~~~ to  return WHILE val:"; val (g,p) = (g,p);
  "~~~~~ to  return GOTO3 val:"; val (g,p) = (g,p);
   val g = pp g;
   (g %|% a) andalso (g %|% b);(*=true*)
"~~~~~ to  return GOTO2 val:"; val (g) = (g);
"~~~~~ to  return GCD_MOD val:"; val (g) = (g);

"----------- fun GCD_MOD---------------------------------";
"----------- fun GCD_MOD --------------------------------";
"----------- fun GCD_MOD --------------------------------";
if GCD_MOD [~18, ~15, ~20, 12, 20, ~13, 2] [8, 28, 22, ~11, ~14, 1, 2] = [~2, ~1, 1] then ()
else error "GCD_MOD [~18, ~15, ~20, 12, 20, ~13, 2] [8, 28, 22, ~11, ~14, 1, 2] = [~2, ~1, 1] changed";

"------------------------------------------------------------";
"--------------------- Multivariate Case --------------------";
"------------------------------------------------------------";

"----------- fun get_coef --------------------------- ---";
"----------- fun get_coef -------------------------------";
"----------- fun get_coef -------------------------------";
if get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 0 = 1 then () else 
error "get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 0 = 1 changed";
if get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 1 = 2 then () else 
error "get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 1 = 2 changed";
if get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 2 = 3 then () else 
error "get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 2 = 3 changed";
if get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 3 = 4 then () else 
error "get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 3 = 4 changed";
if get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 4 = 5 then () else 
error "get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 4 = 5 changed";


"----------- fun get_varlist ----------------------------";
"----------- fun get_varlist ----------------------------";
"----------- fun get_varlist ----------------------------";
if get_varlist [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 0 = [1,1] then () else 
error "get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 0 = [1,1] changed";
if get_varlist [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 1 = [1,2] then () else 
error "get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 1 = [1,2] changed";
if get_varlist [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 2 = [1,3] then () else 
error "get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 2 = [1,3] changed";
if get_varlist [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 3 = [1,4] then () else 
error "get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 3 = [1,4] changed";
if get_varlist [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 4 = [1,5] then () else 
error "get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] 4 = [1,5] changed";

"----------- fun get_coeflist ---------------------------";
"----------- fun get_coeflist ---------------------------";
"----------- fun get_coeflist ---------------------------";
if get_coeflist [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])]  = [1,2,3,4,5] then () else 
error "get_coef [(1,[1,1]),(2,[1,2]),(3,[1,3]),(4,[1,4]),(5,[1,5])] = [1,2,3,4,5] changed";

"----------- fun add_var_to_multipoly -------------------";
"----------- fun add_var_to_multipoly -------------------";
"----------- fun add_var_to_multipoly -------------------";
if add_var_to_multipoly [(1,[1,2]),(2,[2,3])] 0 = [(1, [0, 1, 2]), (2, [0, 2, 3])] then () else 
error "add_var_to_multipoly [(1,[1,2]),(2,[2,3])] 0 = [(1, [0, 1, 2]), (2, [0, 2, 3])] changed";
if add_var_to_multipoly [(1,[1,2]),(2,[2,3])] 1 = [(1, [1, 0,  2]), (2, [2, 0, 3])] then () else 
error "add_var_to_multipoly [(1,[1,2]),(2,[2,3])] 1 = [(1, [1, 0,  2]), (2, [2, 0, 3]) changed";
if add_var_to_multipoly [(1,[1,2]),(2,[2,3])] 2 = [(1, [1, 2, 0]), (2, [2, 3, 0])] then () else 
error "add_var_to_multipoly [(1,[1,2]),(2,[2,3])] 2 = [(1, [1, 2, 0]), (2, [2, 3, 0])] changed";

"----------- fun cero_multipoly -------------------------";
"----------- fun cero_multipoly -------------------------";
"----------- fun cero_multipoly -------------------------";
if cero_multipoly 1 = [(0, [0])] then () else 
error "cero_multipoly 1 = [(0, [0])] changed";
if cero_multipoly 5 = [(0, [0, 0, 0, 0, 0])] then () else 
error "cero_multipoly 1 = [(0, [0, 0, 0, 0, 0])] changed";

"----------- fun short_mv -------------------------------";
"----------- fun short_mv -------------------------------";
"----------- fun short_mv -------------------------------";
"~~~~~ fun short_mv, args:"; val (mvp) = ([(~12, [1]), (2, [1]), (4, [0])]);
"~~~~~ fun short , args:"; val (mvp, new) = (mvp,([]:multipoly));
length mvp =1(*false*);
 get_varlist  mvp 0 = get_varlist  mvp 1; (* true*)
"~~~~~ fun short , args:"; val (mvp, new) = 
([(get_coef  mvp 0 + get_coef  mvp 1,get_varlist mvp 0)] @ nth_drop 0 (nth_drop 0 mvp), new );
length mvp =1(*false*);
get_varlist  mvp 0 = get_varlist mvp 1;(*false*)
"~~~~~ fun short , args:"; val (mvp, new) = ((nth_drop 0 mvp),new @ [nth mvp 0]);
length mvp =1(*true*);
"~~~~~ to  return val:"; val (shortform) = (new @ mvp);

if short_mv [(~3,[0,0]),(4,[0,0]),(3,[1,1]),(~3,[1,1]),(2,[1,2]),(~3,[1,2])] = [(1, [0, 0]), (~1, [1, 2])] 
then () else error "short_mv [(~3,[0,0]),(4,[0,0]),(3,[1,1]),(~3,[1,1]),(2,[1,2]),(~3,[1,2])] = [(1, [0, 0]), (~1, [1, 2])] changed"

"----------- fun greater_var ----------------------------";
"----------- fun greater_var ----------------------------";
"----------- fun greater_var ----------------------------";
if greater_var [3] [4] 
then error " greater_var [3] [4] changed = false"  else ();
if greater_var [1,2] [0,3]
then error " greater_var [1,2] [0,3] changed = false"  else ();
if greater_var  [1,2] [3,0] 
then ()  else error " greater_var  [1,2] [3,0] = true changed";
if greater_var [1,2] [1,2]
then error " greater_var [1,2] [1,2] changed = false"  else ();

"----------- fun order_multipoly ------------------------";
"----------- fun order_multipoly ------------------------";
"----------- fun order_multipoly ------------------------";
if order_multipoly [(3,[1,1]),(2,[1,2]),(~3,[0,0]),(~3,[1,1]),(~3,[1,2]),(4,[0,0])] = [(1, [0, 0]), (~1, [1, 2])] 
then () else error "order_multipoly [(3,[1,1]),(2,[1,2]),(~3,[0,0]),(~3,[1,1]),(~3,[1,2]),(4,[0,0])] = [(1, [0, 0]), (~1, [1, 2])] changed"

"----------- fun lc_multipoly ---------------------------";
"----------- fun lc_multipoly ---------------------------";
"----------- fun lc_multipoly ---------------------------";
if lc_multipoly [(~3,[0,0]),(4,[0,0]),(3,[1,1]),(~3,[1,1]),(2,[1,2]),(~3,[1,2])] = ~1 
then () else error "lc_multipoly [(~3,[0,0]),(4,[0,0]),(3,[1,1]),(~3,[1,1]),(2,[1,2]),(~3,[1,2])] = ~1  changed";
if lc_multipoly [(3,[1,1]),(2,[1,2]),(~3,[0,0])] = 2 
then () else error "lc_multipoly [(3,[1,1]),(2,[1,2]),(~3,[0,0])] = 2   changed";

"----------- fun deg_multipoly -------------------------";
"----------- fun deg_multipoly -------------------------";
"----------- fun deg_multipoly -------------------------";
if deg_multipoly [(~3,[0,0]),(4,[0,0]),(3,[1,1]),(~3,[1,1]),(2,[1,2]),(~3,[1,2])] = [1,2]
then () else error "lc_multipoly [(~3,[0,0]),(4,[0,0]),(3,[1,1]),(~3,[1,1]),(2,[1,2]),(~3,[1,2])] = ~1  changed";
if deg_multipoly [(3,[1,1]),(2,[1,2]),(~3,[0,0])] = [1,2] 
then () else error "lc_multipoly [(3,[1,1]),(2,[1,2]),(~3,[0,0])] = 2   changed";



"----------- fun max_deg_var ----------------------------";
"----------- fun max_deg_var ----------------------------";
"----------- fun max_deg_var ----------------------------";
if max_deg_var [(1,[1,2]),(2,[2,3]),(3,[5,4]),(1,[0,0])] 0 = 5 then ()
else error " max_deg_var [(1,[1,2]),(2,[2,3]),(3,[5,4]),(1,[0,0])] 0 = 5 changed";
if max_deg_var [(1,[1,2]),(2,[2,3]),(3,[5,4]),(1,[0,0])] 1 = 4 then ()
else error " max_deg_var [(1,[1,2]),(2,[2,3]),(3,[5,4]),(1,[0,0])] 1 = 4 changed";

"----------- infix %%/ ----------------------------------";
"----------- infix %%/ ----------------------------------";
"----------- infix %%/ ----------------------------------";
if ([(~3, [2, 0]), (~6, [1, 1]), (~1, [3, 1]),(1, [5, 0]), (2, [4, 1])] %%/ 2) = [(~1, [2, 0]), (~3, [1, 1]), (1, [4, 1])] then ()
else error "[(~3, [2, 0]), (~6, [1, 1]), (~1, [3, 1]),(1, [5, 0]), (2, [4, 1])] %%/ 2  =  [(~1, [2, 0]), (~3, [1, 1]), (1, [4, 1])] changed";

"----------- fun cont -----------------------------------";
"----------- fun cont -----------------------------------";
"----------- fun cont -----------------------------------";
if cont [4,8,12,~2,0,~4] = 2 then () else error " cont [4,8,12,~2,0,~4] = 2 changed";
if cont [3,6,9,19] = 1 then () else error " cont [3,6,9,19] = 1 changed";

"----------- fun cont_multipoly -------------------------";
"----------- fun cont_multipoly -------------------------";
"----------- fun cont_multipoly -------------------------";
if cont_multipoly [(3,[1,2,3,1])] = 3 then () else error " cont_multipoly [(3,[1,2,3,1])] = 3 changed";
if cont_multipoly [(~2, [2, 0]), (4, [5, 0]),(~4, [3, 2]), (2, [2, 3])] = 2 then () 
else error "cont_multipoly [(~2, [2, 0]), (4, [5, 0]),(~4, [3, 2]), (2, [2, 3])] = 2 changed";
if cont_multipoly [(~2, [2, 0]), (5, [5, 0]),(~4, [3, 2]), (2, [2, 3])] = 1 then ()
else error "cont_multipoly [(~2, [2, 0]), (5, [5, 0]),(~4, [3, 2]), (2, [2, 3])] = 1 changed";

"----------- fun evaluate_mv ----------------------------";
"----------- fun evaluate_mv ----------------------------";
"----------- fun evaluate_mv ----------------------------";
if evaluate_mv [(~3, [2, 1]), (2, [0, 1]),(1,[2,0])] 0 1 = [(1, [0]), (~1, [1])]
then () else error " evaluate_mv [(~3, [2, 1]), (2, [0, 1]),(1,[2,0])] 0 1 = [(1, [0]), (~1, [1])] changed";
if evaluate_mv [(~3, [2, 1]), (2, [0, 1]),(1,[2,0])] 1 1 = [(2, [0]), (~2, [2])]
then () else error " evaluate_mv [(~3, [2, 1]), (2, [0, 1]),(1,[2,0])] 1 1 =[(2, [0]), (~2, [2])] changed";
if evaluate_mv [(~3, [2, 1]), (2, [0, 1]),(1,[2,0])] 0 3 = [(9, [0]), (~25, [1])]
then () else error " evaluate_mv [(~3, [2, 1]), (2, [0, 1]),(1,[2,0])] 0 3 = [(9, [0]), (~25, [1])] changed";
if evaluate_mv [(~3, [2, 1]), (2, [0, 1]),(1,[2,0])] 1 3 = [ (6, [0]), (~8, [2])]
then () else error " evaluate_mv [(~3, [2, 1]), (2, [0, 1]),(1,[2,0])] 1 3 = [ (6, [0]), (~8, [2])] changed";

"----------- fun mutli_to_uni ---------------------------";
"----------- fun mutli_to_uni ---------------------------";
"----------- fun mutli_to_uni ---------------------------";
if multi_to_uni [(~3, [1]), (2, [1]), (1, [0])] = [1, ~1]
then () else error " multi_to_uni [(~3, [1]), (2, [1]), (1, [0])] = [1, ~1] changed";

"~~~~~ fun multi_to_uni , args:"; val (mvp) = ([(~3, [0]), (1, [0]), (3, [1]), (~6, [1]),(2,[3])]);
val mvp = order_multipoly mvp;
"~~~~~ fun short, args:"; val (mvp, uvp) = ((short_mv mvp), ([]:unipoly));
length uvp = (nth(get_varlist  mvp 0) 0);(*true*)
"~~~~~ fun short, args:"; val (mvp, uvp) = ((nth_drop 0 mvp), (uvp @ [get_coef mvp 0]));
length uvp = (nth(get_varlist  mvp 0) 0);(*true*)
"~~~~~ fun short, args:"; val (mvp, uvp) = ((nth_drop 0 mvp), (uvp @ [get_coef mvp 0]));
length uvp = (nth(get_varlist  mvp 0) 0);(*false*)
"~~~~~ fun short, args:"; val (mvp, uvp) = (mvp, (uvp @ [0]));
length uvp = (nth(get_varlist  mvp 0) 0);(*true*)
"~~~~~ fun short, args:"; val (mvp, uvp) = ((nth_drop 0 mvp), (uvp @ [get_coef mvp 0]));
"~~~~~ to  return val:"; val (unipoly) = (uvp);



"----------- fun find_new_r  ----------------------------";
"----------- fun find_new_r  ----------------------------";
"----------- fun find_new_r  ----------------------------";
if find_new_r [(2,[0,0]),(1,[1,2])] [(4,[2,2]),(1,[1,2])] ~1 = 1 then ()
else error " find_new_r [(2,[0,0]),(1,[1,2])] [(4,[2,2]),(1,[1,2])] ~1 = 1 changed";
if find_new_r [(2,[0,0]),(1,[1,2])] [(4,[2,2]),(1,[1,2])] 1 = 2 then ()
else error "find_new_r [(2,[0,0]),(1,[1,2])] [(4,[2,2]),(1,[1,2])] 1 = 2 changed";
if find_new_r [(2,[1,2]),(~4,[0,2]),(2,[1,1])] [(1,[1,3]),(~2,[0,3])] 1 = 3 then ()
else error "find_new_r [(2,[1,2]),(~4,[0,2]),(2,[1,1])] [(1,[1,3]),(~2,[0,3])] 1 = 3 changed";

"----------- fun newton_mv  -----------------------------";
"----------- fun newton_mv  -----------------------------";
"----------- fun newton_mv  -----------------------------";
if uni_to_multi [1,2,1,1,3,4] = [(1, [0]), (2, [1]), (1, [2]), (1, [3]), (3, [4]), (4, [5])] then ()
else error "uni_to_multi [1,2,1,1,3,4] = [(1, [0]), (2, [1]), (1, [2]), (1, [3]), (3, [4]), (4, [5])] changed";
if uni_to_multi [1,2,0,0,5,6] = [(1, [0]), (2, [1]), (5, [4]), (6, [5])] then ()
else error "uni_to_multi [1,2,0,0,5,6] = [(1, [0]), (2, [1]), (5, [4]), (6, [5])] changed";

"----------- fun mult_with_new_var  ---------------------";
"----------- fun mult_with_new_var  ---------------------";
"----------- fun mult_with_new_var  ---------------------";
if mult_with_new_var [(3,[0]),(2,[1])] [~1,1] 0 = [(~3, [0, 0]), (3, [1, 0]), (~2, [0, 1]), (2, [1, 1])] then ()
else error "mult_with_new_var [(3,[0]),(2,[1])] [~1,1] 0 = [(~3, [0, 0]), (3, [1, 0]), (~2, [0, 1]), (2, [1, 1])] changed";

"----------- fun greater_deg_multipoly  -----------------";
"----------- fun greater_deg_multipoly  -----------------";
"----------- fun greater_deg_multipoly  -----------------";
greater_deg_multipoly [1,2,3] [1,2,4] =2;

"----------- infix %%+%%  -------------------------------";
"----------- infix %%+%%  -------------------------------";
"----------- infix %%+%%  -------------------------------";
if ([(3,[0,0]),(2,[3,2]),(~3,[1,3])] %%+%% [(3,[3,2]),(~1,[0,0]),(2,[1,1])]) = [(2, [0, 0]), (2, [1, 1]), (5, [3, 2]), (~3, [1, 3])]
then () else error "([(3,[0,0]),(2,[3,2]),(~3,[1,3])] %%+%% [(3,[3,2]),(~1,[0,0]),(2,[1,1])]) = [(2, [0, 0]), (2, [1, 1]), (5, [3, 2]), (~3, [1, 3])] changed";

"----------- infix %%-%%  -------------------------------";
"----------- infix %%-%%  -------------------------------";
"----------- infix %%-%%  -------------------------------";
if ([(2, [0, 0]), (2, [1, 1]), (5, [3, 2]), (~3, [1, 3])] %%-%% [(3,[3,2]),(~1,[0,0]),(2,[1,1])]) = [(3, [0, 0]), (2, [3, 2]), (~3, [1, 3])]
then () else error "([(2, [0, 0]), (2, [1, 1]), (5, [3, 2]), (~3, [1, 3])] %%-%% [(3,[3,2]),(~1,[0,0]),(2,[1,1])]) = [(3, [0, 0]), (2, [3, 2]), (~3, [1, 3])] changed";

"----------- infix %%*%%  -------------------------------";
"----------- infix %%*%%  -------------------------------";
"----------- infix %%*%%  -------------------------------";
if ([(~1,[0]),(1,[1])] %%*%% [(~2,[0]),(1,[1])]) = [(2, [0]), (~3, [1]), (1, [2])]
then () else error "([(~1,[0]),(1,[1])] %%*%% [(~2,[0]),(1,[1])]) = [(2, [0]), (~3, [1]), (1, [2])] changed";
if ([(~2,[0]),(1,[1])] %%*%% [(1,[1]),(2,[0])]) = [(~4, [0]), (1, [2])] then ()
else error "([(~2,[0]),(1,[1])] %%*%% [(1,[1]),(2,[0])]) = [(~4, [0]), (1, [2])] changed";
if ([(3,[0,0]),(2,[0,1])] %%*%% [(1,[1,0]),(~1,[0,0])]) = mult_with_new_var [(3,[0]),(2,[1])] [~1,1] 0
then () else error 
"([(3,[0,0]),(2,[0,1])] %%*%% [(1,[1,0]),(~1,[0,0])]) = mult_with_new_var [(3,[0]),(2,[1])] [~1,1] 0 changed";

"----------- fun newton_first  --------------------------";
"----------- fun newton_first  --------------------------";
"----------- fun newton_first  --------------------------";
if newton_first [1, 2] [[(1,[1,1])], [(4,[1,1])]] 1 = 
([(~2, [1, 0, 1]), (3, [1, 1, 1])], [~1, 1], [[(3, [1, 1])]]) 
then () else error 
"newton_first [1, 2] [[(1,[1,1])], [(4,[1,1])]] 1 = "
"([(~2, [1, 0, 1]), (3, [1, 1, 1])], [~1, 1], [[(3, [1, 1])]]) changed";

"----------- fun newton_mv  -----------------------------";
"----------- fun newton_mv  -----------------------------";
"----------- fun newton_mv  -----------------------------";
if newton_mv [1, 2 ] [[(1,[1,1])], [(4,[1,1])]] [] [1] (cero_multipoly 2) 1 = 
([(~2, [1, 0, 1]), (3, [1, 1, 1])], [~1, 1], [[(3, [1, 1])]])
then () else error
"newton_mv [1, 2 ] [[(1,[1,1])], [(4,[1,1])]] [] [1] (cero_multipoly 2) 1 = "
"([(~2, [1, 0, 1]), (3, [1, 1, 1])], [~1, 1], [[(3, [1, 1])]]) changed";
if newton_mv [1, 2, 3 ] [[(4,[1,1])], [(9,[1,1])]] [[(3, [1, 1])]] [~1, 1] [(~2, [1, 0, 1]), (3, [1, 1, 1])] 1 = 
([(1, [1, 2, 1])], [2, ~3, 1], [[(5, [1, 1])], [(1, [1, 1])]])
then () else error
"newton_mv [1, 2, 3 ] [[(4,[1,1])], [(9,[1,1])]] [[(3, [1, 1])]] [~1, 1] [(~2, [1, 0, 1]), (3, [1, 1, 1])] 1 ="
" ([(1, [1, 2, 1])], [2, ~3, 1], [[(5, [1, 1])], [(1, [1, 1])]]) changed";
 

"~~~~~ fun newton_mv, args:"; val (x,f,steps,t,p,order) = ([1, 2, 3, 4],[[(9, [0]), (5, [1])], [(16, [0]), (7, [1])]], [[(5, [0]), (2, [1])], [(1, [0])]], [2, ~3, 1], [(1, [2, 0]), (~1, [0, 1]), (2, [1, 1])], 0  );
length x = 2; (* false *)
val new_value_poly = multi_to_uni((uni_to_multi t)  %%*%% (uni_to_multi [(nth x (length x - 2) )* ~1, 1]));
val new_steps = [((nth f (length f - 1)) %%/ ((nth x (length x - 1)) - (nth x (length x - 2)))) %%-%% ((nth f (length f - 2)))];

"~~~~~ fun next_step, args:"; val (steps,new_steps, x') = (steps, new_steps, x);
steps = []; (*false*)

"~~~~~ fun next_step, args:"; val (steps,new_steps, x') = (nth_drop 0 steps, new_steps @ [(((nth new_steps (length new_steps - 1)) %%-%%(nth steps 0))) %%/
                                  ((nth x' (length x' - 1)) - (nth x' (length x' - 3)))], nth_drop (length x' - 2) x');steps = []; (*false*)

"~~~~~ fun next_step, args:"; val (steps,new_steps, x') = (nth_drop 0 steps, new_steps @ [(((nth new_steps (length new_steps - 1)) %%-%%(nth steps 0))) %%/
                                  ((nth x' (length x' - 1)) - (nth x' (length x' - 3)))], nth_drop (length x' - 2) x');
steps = []; (*true*)
val steps = new_steps;
val polynom' =  p %%+%% (mult_with_new_var (nth steps (length steps - 1)) new_value_poly order);

"----------- fun listgreater  ---------------------------";
"----------- fun listgreater  ---------------------------";
"----------- fun listgreater  ---------------------------";
if listgreater  [1,2,3,4,5] [1,2,3,4,5] = true then ()
else error " listgreater  [1,2,3,4,5] [1,2,3,4,5] = true changed";
if listgreater [1,2,3,4] [1,2,3,5] = false then () 
else error "listgreater [1,2,3,4] [1,2,3,5] = false changed"  ;
if listgreater [1,4,5,4] [0,3,4,5] = false then ()
else error "listgreater [1,2,3,4] [0,3,4,5] = false changed ";

"----------- fun greater_deg_multipoly  -----------------";
"----------- fun greater_deg_multipoly  -----------------";
"----------- fun greater_deg_multipoly  -----------------";
if greater_deg_multipoly  [1,2,3,4,5] [1,2,3,4,5] = 0 then ()
else error " greater_deg_multipoly  [1,2,3,4,5] [1,2,3,4,5] = 0 changed";
if greater_deg_multipoly [1,2,3,4] [5,2,8,1] = 1 then () 
else error "greater_deg_multipoly [1,2,3,4] [1,2,3,5] = 1 changed"  ;
if greater_deg_multipoly [1,4,5,4] [0,3,4,5] = 2 then ()
else error "greater_deg_multipoly [1,2,3,4] [0,3,4,5] = 2 changed ";

"----------- finfix  %%|%%  ------------------------------";
"----------- finfix  %%|%%  ------------------------------";
"----------- finfix  %%|%%  ------------------------------";
if [(1, [0, 0]), (~1, [0, 1])] %%|%% [(1, [0, 0]), (~1, [0, 1])] then () 
else error "[(1, [0, 0]), (~1, [0, 1])] %%|%% [(1, [0, 0]), (~1, [0, 1])] = true changed";
if [(3,[1,0])] %%|%% [(9,[1,1]),(12,[2,1]),(~3,[1,2])] then ()
else error "[(3,[1,0])] %%|%% [(9,[1,1]),(12,[2,1]),(~3,[1,2])] = true changed";
if [(3,[2,1])] %%|%% [(9,[1,1]),(12,[2,1]),(~3,[1,2])] 
then error "[(3,[2,1])] %%|%% [(9,[1,1]),(12,[2,1]),(~3,[1,2])] = false changed" else ();

"----------- fun GCD_MODm  ------------------------------";
"----------- fun GCD_MODm  ------------------------------";
"----------- fun GCD_MODm  ------------------------------";

if GCD_MODm 
  [(~3,[2,0]),(1,[5,0]),(3,[0,1]),(~6,[1,1]),(~1,[3,1]),(2,[4,1]),(1,[3,2]),(~1,[1,3]),(2,[2,3])]
  [(2,[2,0]),(~2,[0,1]),(4,[1,1]),(~1,[3,1]),(1,[1,2]),(~1,[2,2]),(~1,[0,3]),(2,[1,3])] 2 1 0 
  = [(1, [2, 0]), (~1, [0, 1]), (2, [1, 1])] then () else error 
"GCD_MODm [(~3,[2,0]),(1,[5,0]),(3,[0,1]),(~6,[1,1]),(~1,[3,1]),(2,[4,1]),(1,[3,2]),(~1,[1,3]),(2,[2,3])]"
"[(2,[2,0]),(~2,[0,1]),(4,[1,1]),(~1,[3,1]),(1,[1,2]),(~1,[2,2]),(~1,[0,3]),(2,[1,3])] 2 1 0 "
"= [(1, [2, 0]), (~1, [0, 1]), (2, [1, 1])] changed";
(* -xy +xy^2z+yz - 1*)(* xy +1*) (*=*) (*xy - 1*)
if GCD_MODm [(~1,[0,0,0]),(1,[0,1,1]),(1,[1,2,1]),(~1,[1,1,0])] [(1,[0,0,0]),(1,[1,1,0])] 3 2 0 
= [(1, [0, 0, 0]), (1, [1, 1, 0])] then () else error
"GCD_MODm [(~1,[0,0,0]),(1,[0,1,1]),(1,[1,2,1]),(~1,[1,1,0])] [(1,[0,0,0]),(1,[1,1,0])] 3 2 0 "
"= [(1, [0, 0, 0]), (1, [1, 1, 0])] changed";

" ~~~ 1 ~~~";
"~~~~~ fun GCD_MODm , args:"; val (a,b,n,s,r) = ([(1,[1,2,1])], [(1,[1,2,1])], 3, 2, 0);
s = 0; (*false*)
val M = 1 +  Int.min (max_deg_var a (length(get_varlist a 0)- 2), max_deg_var b (length(get_varlist b 0)- 2)); 
 " ~~~ 1 ~~~";
"~~~~~ fun GOTO2 , args:"; val(a, b, n, s, M, r_list, steps ) =(a, b, n, s, M, [], []);
val r = find_new_r a b r;
val r_list = r_list @ [r];
val (a',b',n',s',r',r_list',steps',M') = ( a,b,n,s,r,r_list,steps,M);
(*g_1=*)
 " ~~~ 1.1 ~~~";
"~~~~~ fun GCD_MODm , args:"; val (a,b,n,s,r) = ((order_multipoly (evaluate_mv a (s- 1) r)), ( order_multipoly (evaluate_mv b (s- 1) r)), (n- 1), (s- 1), 0);
s = 0; (*false*)

val M = 1 +  Int.min (max_deg_var a (length(get_varlist a 0)- 2), max_deg_var b (length(get_varlist b 0)- 2)); 
 " ~~~ 1.1 ~~~";
"~~~~~ fun GOTO2 , args:"; val(a, b, n, s, M, r_list, steps ) =(a, b, n, s, M, [], []);
val r = find_new_r a b r;
val r_list = r_list @ [r];
val (a'',b'',n'',s'',r'',r_list'') = ( a,b,n,s,r,r_list);
(*g_2=*)
 " ~~~ 1.1.1 ~~~";
"~~~~~ fun GCD_MODm , args:"; val (a,b,n,s,r) = ((order_multipoly (evaluate_mv a (s- 1) r)), ( order_multipoly (evaluate_mv b (s- 1) r)), (n- 1), (s- 1), 0);
s = 0; (*true*)
val g_2 =uni_to_multi((GCD_MOD (pp(multi_to_uni a)) (pp(multi_to_uni b))) %* (abs (Integer.gcd (cont (multi_to_uni a)) (cont( multi_to_uni b)))))
val ( a,b,n,s,r,r_list) = (a'',b'',n'',s'',r'',r_list'');
val g_r = g_2;
val steps = steps @ [g_r];
 " ~~~ 1.1 ~~~";
"~~~~~ fun GOTO3 , args:"; val (a, b, n, s, M, g_r, r, r_list, steps) =
(a, b, n, s, M, g_r, r, r_list, steps);
val m = 1;
 " ~~~ 1.1.W1 ~~~";
"~~~~~ fun WHILE , args:"; val (a ,b ,n ,s, M, m, r, r_list, steps, g, g_n ,mult) =
(a, b, n, s, M, m,r, r_list, steps, g_r, ( cero_multipoly (s+1)), [1]);
; (*false*)
val r = find_new_r a b r; 
val r_list = r_list @ [r];
val (a'',b'',n'',s'',r'',r_list'') = ( a,b,n,s,r,r_list);
(*g_3=*)
 " ~~~ 1.1.W1.1 ~~~";
"~~~~~ fun GCD_MODm , args:"; val (a,b,n,s,r) = ((order_multipoly (evaluate_mv a (s- 1) r)), ( order_multipoly (evaluate_mv b (s- 1) r)), (n- 1), (s- 1), 0);
s = 0; (*true*)
val g_3 =uni_to_multi((GCD_MOD (pp(multi_to_uni a)) (pp(multi_to_uni b))) %* (abs (Integer.gcd (cont (multi_to_uni a)) (cont( multi_to_uni b)))));
 " ~~~ 1.1.W1 ~~~";
val ( a,b,n,s,r,r_list) = (a'',b'',n'',s'',r'',r_list'');
val g_r = g_3;
greater_var  (deg_multipoly g) (deg_multipoly g_r) (*false*);
(deg_multipoly g)= (deg_multipoly g_r) (*true*);
val (g_n, new, steps) = newton_mv r_list [g, g_r] steps mult g_n (s- 1);
if g_n = [(1,[1,1])] then () else error "g_n is false" ;
(nth steps (length steps - 1)) = (cero_multipoly s) (* false*);
 " ~~~ 1.1.W2 ~~~";
"~~~~~ fun WHILE , args:"; val (a ,b ,n ,s, M, m, r, r_list, steps, g, g_n ,mult) =
(a, b, n, s, M, (m+1), r, r_list, steps, g_r, g_n, new);
m > M; (*false*)
val r = find_new_r a b r; 
val r_list = r_list @ [r];
val (a'',b'',n'',s'',r'',r_list'') = ( a,b,n,s,r,r_list);
(*g_4=*)
" ~~~ 1.1.W2.1 ~~~";
"~~~~~ fun GCD_MODm , args:"; val (a,b,n,s,r) = ((order_multipoly (evaluate_mv a (s- 1) r)), ( order_multipoly (evaluate_mv b (s- 1) r)), (n- 1), (s- 1), 0);
s = 0; (*true*)
val g_r =uni_to_multi((GCD_MOD (pp(multi_to_uni a)) (pp(multi_to_uni b))) %* (abs (Integer.gcd (cont (multi_to_uni a)) (cont( multi_to_uni b)))));
val ( a,b,n,s,r,r_list) = (a'',b'',n'',s'',r'',r_list'');
" ~~~ 1.1.W2 ~~~";
greater_var  (deg_multipoly g) (deg_multipoly g_r) (*false*);
(deg_multipoly g)= (deg_multipoly g_r) (*true*);
val (g_n, new, steps) = newton_mv r_list [g, g_r] steps mult g_n (s- 1);
if g_n = [(1,[1,1])] then () else error "g_n is false" ;
(nth steps (length steps - 1)) = (cero_multipoly s) (* true*);
 " ~~~ 1.1.W3 ~~~";
"~~~~~ fun WHILE , args:"; val (a ,b ,n ,s, M, m, r, r_list, steps, g, g_n ,mult) =
(a, b, n, s, M, (M+1), r, r_list, steps, g_r, g_n, new);
m > M; (*true*)
g_n %%|%% a andalso g_n %%|%% b(*true*);
 " ~~~ 1 ~~~";
val ( a,b,n,s,r,r_list,steps,M) = (a',b',n',s',r',r_list',steps',M');
val g_1=g_n;
val g_r = g_1;
val steps = steps @ [g_r];
 " ~~~ 1 ~~~";
"~~~~~ fun GOTO3, args:"; val (a, b, n, s, M, g_r, r, r_list, steps) = 
(a, b, n, s, M, g_r, r, r_list, steps);
val m = 1;
 " ~~~ 1.W1 ~~~";
"~~~~~ fun WHILE, args:"; val (a, b, n, s, M, m, r, r_list, steps ,g, g_n, mult) = 
(a, b, n, s, M, m, r, r_list, steps ,g_r, ( cero_multipoly (s+1)), [1]);
val g' = g;
m > M; (*false*)
val r = find_new_r a b r; 
val r_list = r_list @ [r];
val (a',b',n',s',r',r_list',steps',m') = ( a,b,n,s,r,r_list,steps,m);
(* g_2 = GCD_MODm *)
 " ~~~ 1.W1.1 ~~~";
"~~~~~ fun GCD_MODm, args:"; val  (a,b,n,s,r) = 
((order_multipoly  (evaluate_mv a (s- 1) r)),(order_multipoly  (evaluate_mv b (s- 1) r)), (n- 1), (s- 1), 0);
s = 0; (*false*)
val M = 1 +  Int.min(max_deg_var a (length(get_varlist a 0)- 2), max_deg_var b (length(get_varlist b 0)- 2)); 
 " ~~~ 1.W1.1 ~~~";
"~~~~~ fun GOTO2, args:"; val (a,b,n,s,M,r_list,steps) =
(a, b, n, s, M, [], []);
val r = find_new_r a b r;
val r_list = r_list @ [r];
val (a'',b'',n'',s'',r'',r_list'') = ( a,b,n,s,r,r_list);
 " ~~~ 1.W1.1.1 ~~~";
 (*g_r=*)
"~~~~~ fun GCD_MODm, args:"; val  (a,b,n,s,r) = 
((order_multipoly  (evaluate_mv a (s- 1) r)),(order_multipoly  (evaluate_mv b (s- 1) r)), (n- 1), (s- 1), 0);
s = 0; (*true*)
val g_r = uni_to_multi((GCD_MOD (pp(multi_to_uni a)) (pp(multi_to_uni b))) %* (abs (Integer.gcd (cont (multi_to_uni a)) (cont( multi_to_uni b)))));
 " ~~~ 1.W1.1 ~~~";
val ( a,b,n,s,r,r_list) = (a'',b'',n'',s'',r'',r_list'');
val steps = steps @ [g_r];
 " ~~~ 1.W1.1 ~~~";
"~~~~~ fun GOTO3, args:"; val (a, b, n, s, M, g_r, r, r_list, steps) = 
(a, b, n, s, M, g_r, r, r_list, steps);
val m = 1;
val (g'',g_n'',g_r'', m'',new'') = (g,g_n,g_r,m,mult);
 " ~~~ 1.W1.1.W1 ~~~";
"~~~~~ fun WHILE, args:"; val (a, b, n, s, M, m, r, r_list, steps ,g, g_n, mult) = 
(a, b, n, s, M, m, r, r_list, steps ,g_r, ( cero_multipoly (s+1)), [1]);

m > M; (*false*)
val r = find_new_r a b r; 
val r_list = r_list @ [r];
val (a'',b'',n'',s'',r'',r_list'') = ( a,b,n,s,r,r_list);
(* g_r = GCD_MODm *)
 " ~~~ 1.W1.1.W1.1 ~~~";
 (*g_r=*)
"~~~~~ fun GCD_MODm, args:"; val  (a,b,n,s,r) = 
((order_multipoly  (evaluate_mv a (s- 1) r)),(order_multipoly  (evaluate_mv b (s- 1) r)), (n- 1), (s- 1), 0);
s = 0; (*true*)
val g_r = uni_to_multi((GCD_MOD (pp(multi_to_uni a)) (pp(multi_to_uni b))) %* (abs (Integer.gcd (cont (multi_to_uni a)) (cont( multi_to_uni b)))));
 " ~~~ 1.W1.1.W1 ~~~";
val ( a,b,n,s,r,r_list) = (a'',b'',n'',s'',r'',r_list'');
greater_var  (deg_multipoly g) (deg_multipoly g_r); (*false*)
(deg_multipoly g)= (deg_multipoly g_r); (*true*)
val (g_n, new, steps) = newton_mv r_list [g, g_r] steps mult g_n (s- 1);
(nth steps (length steps - 1)) = (cero_multipoly s); (*false*)
 " ~~~ 1.W1.1.W2 ~~~";
"~~~~~ fun WHILE, args:"; val (a, b, n, s, M, m, r, r_list, steps ,g, g_n, mult) = 
(a, b, n, s, M, m+1, r, r_list, steps ,g_r, g_n,new);
m > M; (*false*)
val r = find_new_r a b r; 
val r_list = r_list @ [r];
val (a'',b'',n'',s'',r'',r_list'') = ( a,b,n,s,r,r_list);
(* g_r = GCD_MODm *)
 " ~~~ 1.W1.1.W2.1 ~~~";
 (*g_r=*)
"~~~~~ fun GCD_MODm, args:"; val  (a,b,n,s,r) = 
((order_multipoly  (evaluate_mv a (s- 1) r)),(order_multipoly  (evaluate_mv b (s- 1) r)), (n- 1), (s- 1), 0);
s = 0; (*true*)
val g_r = uni_to_multi((GCD_MOD (pp(multi_to_uni a)) (pp(multi_to_uni b))) %* (abs (Integer.gcd (cont (multi_to_uni a)) (cont( multi_to_uni b)))));
 " ~~~ 1.W1.1.W2 ~~~";
val ( a,b,n,s,r,r_list) = (a'',b'',n'',s'',r'',r_list'');
greater_var  (deg_multipoly g) (deg_multipoly g_r); (*false*)
(deg_multipoly g)= (deg_multipoly g_r); (*true*)
val (g_n, new, steps) = newton_mv r_list [g, g_r] steps mult g_n (s- 1);
(nth steps (length steps - 1)) = (cero_multipoly s); (*true*)
 " ~~~ 1.W1.1.W3 ~~~";
"~~~~~ fun WHILE, args:"; val (a, b, n, s, M, m, r, r_list, steps ,g, g_n, mult) = 
(a, b, n, s, M, M+1, r, r_list, steps ,g_r, g_n,new);
m>M; (*true*)
g_n %%|%% a andalso g_n %%|%% b; (*true*)
 " ~~~ 1.W1.1 ~~~";
val ( a,b,n,s,r,r_list,g,M) = (a',b',n',s',r',r_list',g',M');
val g_r = g_n;
greater_var  (deg_multipoly g) (deg_multipoly g_r); (*false*)
(deg_multipoly g)= (deg_multipoly g_r); (*true*)
val (g_n,steps,m,new) = (g_n'',steps',m'',new'');
 " ~~~ 1.W2 ~~~";
val (g_n, new, steps) = newton_mv r_list [g, g_r] steps new g_n (s- 1);
(nth steps (length steps - 1)) = (cero_multipoly s); (*false*)
"~~~~~ fun WHILE, args:"; val (a, b, n, s, M, m, r, r_list, steps ,g, g_n, mult) = 
(a, b, n, s, M, m+1, r, r_list, steps ,g_r, g_n,new); 
m > M; (*false*)
val r = find_new_r a b r; 
val r_list = r_list @ [r];
val (a',b',n',s',r',r_list',steps',m',M',g') = ( a,b,n,s,r,r_list,steps,m,M,g);
(* g_2 = GCD_MODm *)
 " ~~~ 1.W2.1 ~~~";
"~~~~~ fun GCD_MODm, args:"; val  (a,b,n,s,r) = 
((order_multipoly  (evaluate_mv a (s- 1) r)),(order_multipoly  (evaluate_mv b (s- 1) r)), (n- 1), (s- 1), 0);
s = 0; (*false*)
val M = 1 +  Int.min(max_deg_var a (length(get_varlist a 0)- 2), max_deg_var b (length(get_varlist b 0)- 2)); 
 " ~~~ 1.W2.1 ~~~";
"~~~~~ fun GOTO2, args:"; val (a,b,n,s,M,r_list,steps) =
(a, b, n, s, M, [], []);
val r = find_new_r a b r;
val r_list = r_list @ [r];
val (a'',b'',n'',s'',r'',r_list'') = ( a,b,n,s,r,r_list);
 " ~~~ 1.W2.1.1 ~~~";
 (*g_r=*)
"~~~~~ fun GCD_MODm, args:"; val  (a,b,n,s,r) = 
((order_multipoly  (evaluate_mv a (s- 1) r)),(order_multipoly  (evaluate_mv b (s- 1) r)), (n- 1), (s- 1), 0);
s = 0; (*true*)
val g_r = uni_to_multi((GCD_MOD (pp(multi_to_uni a)) (pp(multi_to_uni b))) %* (abs (Integer.gcd (cont (multi_to_uni a)) (cont( multi_to_uni b)))));
 " ~~~ 1.W2.1 ~~~";
val ( a,b,n,s,r,r_list) = (a'',b'',n'',s'',r'',r_list'');
val steps = steps @ [g_r];
 " ~~~ 1.W2.1 ~~~";
"~~~~~ fun GOTO3, args:"; val (a, b, n, s, M, g_r, r, r_list, steps) = 
(a, b, n, s, M, g_r, r, r_list, steps);
val m = 1;
val (g'',g_n'',g_r'', m'',new'') = (g,g_n,g_r,m,mult);
 " ~~~ 1.W2.1.W1 ~~~";
"~~~~~ fun WHILE, args:"; val (a, b, n, s, M, m, r, r_list, steps ,g, g_n, mult) = 
(a, b, n, s, M, m, r, r_list, steps ,g_r, ( cero_multipoly (s+1)), [1]);
m > M; (*false*)
val r = find_new_r a b r; 
val r_list = r_list @ [r];
val (a'',b'',n'',s'',r'',r_list'') = ( a,b,n,s,r,r_list);
(* g_r = GCD_MODm *)
 " ~~~ 1.W2.1.W1.1 ~~~";
 (*g_r=*)
"~~~~~ fun GCD_MODm, args:"; val  (a,b,n,s,r) = 
((order_multipoly  (evaluate_mv a (s- 1) r)),(order_multipoly  (evaluate_mv b (s- 1) r)), (n- 1), (s- 1), 0);
s = 0; (*true*)
val g_r = uni_to_multi((GCD_MOD (pp(multi_to_uni a)) (pp(multi_to_uni b))) %* (abs (Integer.gcd (cont (multi_to_uni a)) (cont( multi_to_uni b)))));
 " ~~~ 1.W2.1.W1 ~~~";
val ( a,b,n,s,r,r_list) = (a'',b'',n'',s'',r'',r_list'');
greater_var  (deg_multipoly g) (deg_multipoly g_r); (*false*)
(deg_multipoly g)= (deg_multipoly g_r); (*true*)
val (g_n, new, steps) = newton_mv r_list [g, g_r] steps mult g_n (s- 1);
(nth steps (length steps - 1)) = (cero_multipoly s); (*false*)
 " ~~~ 1.W2.1.W2 ~~~";
"~~~~~ fun WHILE, args:"; val (a, b, n, s, M, m, r, r_list, steps ,g, g_n, mult) = 
(a, b, n, s, M, m+1, r, r_list, steps ,g_r, g_n,new);
m > M; (*false*)
val r = find_new_r a b r; 
val r_list = r_list @ [r];
val (a'',b'',n'',s'',r'',r_list'') = ( a,b,n,s,r,r_list);
(* g_r = GCD_MODm *)
 " ~~~ 1.W2.1.W2.1 ~~~";
 (*g_r=*)
"~~~~~ fun GCD_MODm, args:"; val  (a,b,n,s,r) = 
((order_multipoly  (evaluate_mv a (s- 1) r)),(order_multipoly  (evaluate_mv b (s- 1) r)), (n- 1), (s- 1), 0);
s = 0; (*true*)
val g_r = uni_to_multi((GCD_MOD (pp(multi_to_uni a)) (pp(multi_to_uni b))) %* (abs (Integer.gcd (cont (multi_to_uni a)) (cont( multi_to_uni b)))));
 " ~~~ 1.W2.1.W2 ~~~";
val ( a,b,n,s,r,r_list) = (a'',b'',n'',s'',r'',r_list'');
greater_var  (deg_multipoly g) (deg_multipoly g_r); (*false*)
(deg_multipoly g)= (deg_multipoly g_r); (*true*)
val (g_n, new, steps) = newton_mv r_list [g, g_r] steps mult g_n (s- 1);
(nth steps (length steps - 1)) = (cero_multipoly s); (*true*)
 " ~~~ 1.W2.1.W3 ~~~";
"~~~~~ fun WHILE, args:"; val (a, b, n, s, M, m, r, r_list, steps ,g, g_n, mult) = 
(a, b, n, s, M, M+1, r, r_list, steps ,g_r, g_n,new);
m>M; (*true*)
g_n %%|%% a andalso g_n %%|%% b; (*true*)
 " ~~~ 1.W2.1 ~~~";
val ( a,b,n,s,r,r_list,g) = (a',b',n',s',r',r_list',g');
val g_r = g_n;
greater_var  (deg_multipoly g) (deg_multipoly g_r); (*false*)
(deg_multipoly g)= (deg_multipoly g_r); (*true*)
val (g_n,steps,m,mult) = (g_n'',steps',m'',new'');
 " ~~~ 1.W2 ~~~";
 val (g_n, new, steps) = newton_mv r_list [g, g_r] steps mult g_n (s- 1);
"~~~~~ to  return val:"; val (g) = (g_n);



if g =  [(1, [1, 2, 1])] then () else error "GCD_MODm [(1, [1, 2, 1])] [(1, [1, 2, 1])] has changes.";

" ==========================  END  ========================== ";
  fun nth _ []      = error "nth _ []" (*Isabelle2002, still saved the hours of update*)
    | nth 1 (x::_) = x
    | nth n (_::xs) = nth (n- 1) xs;
(*fun nth xs i = List.nth (xs, i);       recent Isabelle: TODO update all isac code   *)

