(* Title:  BaseDefinitions/references.sml
   Author: Walther Neuper
   (c) due to copyright terms

References into Know_Store.
*)
signature REFERENCES =
sig
  type T
  val empty: T
  val to_string: T -> string
(*val for_eval: T -> T -> T*)
  val select_input: T -> T -> T

  type input
  type input_pos
  val parse_input_pos: input_pos parser
  type input_id
  val thy_empty_input: ThyC.id
  val id_empty_input: input_id
  val empty_input: input
  val explode_id: input_id -> References_Def.id
  val implode_id: References_Def.id -> input_id
  val select_from_input: input -> T -> T -> T

  val explode: input -> T
  val implode: T ->  input
  val for_template: input -> T -> input

  val complete: Ctree.state -> Ctree.state
  val are_complete: Ctree.state -> bool
end

(**)
structure References(**): REFERENCES(**) =
struct
(**)

type id = References_Def.id;
type T = References_Def.T;
val empty = References_Def.empty;
val to_string = References_Def.to_string;
val select_input = References_Def.select_input;

type input_id = string
type input = ThyC.id * input_id * input_id
val thy_empty_input = "__"
val id_empty_input = "__/__"
val empty_input = (thy_empty_input, id_empty_input, id_empty_input)
val explode_id = References_Def.explode_id
val implode_id = References_Def.implode_id
fun select_from_input (input_thy, input_pbl, input_met) (o_thy, o_pbl, o_met) (thy, pbl, met) = 
  (if input_thy = thy_empty_input then if thy = ThyC.id_empty then o_thy else thy
   else input_thy,
   if input_pbl = id_empty_input then if pbl = Problem.id_empty then o_pbl else pbl
   else explode_id input_pbl,
   if input_met = id_empty_input then if met = MethodC.id_empty then o_met else met
   else explode_id input_met)

fun explode (input_thy, input_pbl, input_met) =
  (input_thy, References_Def.explode_id input_pbl, References_Def.explode_id input_met)
fun implode (thy_id, pbl_id, met_id) =
  (thy_id, implode_id pbl_id, implode_id met_id)

type input_pos = ((string * Position.T) * ((string * Position.T) * (string * Position.T)))
val parse_input_pos  =
      \<^keyword>\<open>Theory_Ref\<close> -- \<^keyword>\<open>:\<close> |-- Parse.!!! (Parse.position Parse.name) --
    ( \<^keyword>\<open>Problem_Ref\<close> -- \<^keyword>\<open>:\<close> |-- Parse.!!! (Parse.position Parse.name) --
    ( \<^keyword>\<open>Method_Ref\<close> -- \<^keyword>\<open>:\<close> |-- Parse.!!! (Parse.position Parse.name)))

(* decide which references are displayed by the Specification-template *)
fun for_template (input_thy, input_pbl, input_met) (store_thy, store_pbl, store_met) =
  (if input_thy = thy_empty_input then input_thy else store_thy,
  if input_pbl = id_empty_input then input_pbl else implode_id store_pbl,
  if input_met = id_empty_input then input_met else implode_id store_met)

(* have (thy, pbl, met) _all_ been specified explicitly ? *)
fun are_complete (pt, pos as (p, _)) = 
  if (not o Ctree.is_pblobj o (Ctree.get_obj I pt)) p then
    raise ERROR ("References.is_complete: called by PrfObj at " ^ Pos.pos'2str pos)
  else
    let
      val (dI, pI, mI) = Ctree.get_obj Ctree.g_spec pt p
    in
      dI <> ThyC.id_empty andalso pI <> Problem.id_empty andalso mI <> MethodC.id_empty
    end;

(* complete References and save them to the state *)
fun complete (pt, pos as (p, _)) = 
  let
    val (ospec, spec) = case Ctree.get_obj I pt p of
      Ctree.PblObj {origin = (_, ospec, _), spec, ...} => (ospec, spec)
    | _ => raise ERROR "References.complete: uncovered case get_obj"
    val pt = Ctree.update_spec pt p (select_input ospec spec)
  in
    (pt, pos)
  end

(**)end(**)
